'use strict';

// This module handles the OpenLayers-based map

/* global appCore,IMatchModals,CameraDatabase,MapUtility,moment,LatLon */

function OLMapModule() {
    
    var self = this;

    var olMap = null;
    
    var vectorSource = null;
    var clusterSource = null;
    var fileLayer = null;
    var dataLayer = null;

    // The layer with the target marker etc.
    var markerLayer = null;
    var markerSource = null;
    var markerTarget = null;
    var focusedMarker = null;

    // For lines connecting source and destination markers
    var lineSource = null;
    var lineLayer = null;

    // Locations
    var locationLayer = null;
    var locationSource = null;

    // For tracks
    var trackLayer = null;
    var trackSource = null;
    var trackMarkerLayer = null;
    var trackMarkerDataLayer = null;
    var trackMarkerSource = null;

    // File route/track
    var showFileRoute = false;
    var fileRouteLayer = null;
    var fileRouteSource = null;

    var eatLeftClick = false;
    // To process drag & drop of features
    var diTarget = null;
    var diFiles = null;
    // We use this with the 'change' event of each feature to 
    // keep track of the last modified file feature. We need that
    // in the modifyend handler
    var diFilesLMF = null;
    var diLocations = null;
    var diLocationsLMF = null;

    var styleCache = {};

    var files = [];

    var locations = [];

    const TRACK_COLORS_COUNT = 5;
    const TRACK_COLORS = ["rgba(0, 0, 255, 0.5)", "rgba(0, 73, 124, 1)", "rgba(0, 98, 168, 1)", "rgba(0, 124, 211, 1)", "rgba(0, 149, 255, 1)"];

    // The popup we use for showing information
    self.filePopup = {
        // Links to the div on the map
        popup: null,

        // The current popup coordinate (where the user clicked, not necessarily the same as the feature coordinate!)
        coordinate: null,

        // The last used feature
        feature: null,

        // The index of the current file (when we have a cluster of files)
        index: 0,
        
        // The number of files in the cluster
        files: 0,

        // The file id
        fileId: 0
    }

    self.targetPopup = {
        // Links to the div on the map
        popup: null,
    }

    self.locationPopup = {
        feature: null,
        popup: null,
    }

    self.popup = null;

    // State of the select tool
    
    var selectTool = {

        object : null,

        iaModify: null,
        iaTranslate: null,
        
        objectType: '',

        source : null,

        layer: null,

        style : new ol.style.Style({

            stroke: new ol.style.Stroke({
                width: 4,
                color: 'rgba(216,28,38,0.8)'
            }),

            fill: new ol.style.Fill({
                color: 'rgba(50,160,204,0.35)',
            })
        })
    }

    // In this we manage all FOV polygons visible on the map.
    // This structure holds the polgons wrapped in features and
    // controls the styles.

    var fovLayer = {
        source : null,
        layer: null,
        style : new ol.style.Style({

            stroke: new ol.style.Stroke({
                width: 4,
                color: 'rgba(0,0,255,0.4)'
            }),

            fill: new ol.style.Fill({
                color: 'rgba(0,0,255,0.25)',
            })
        })
    }


    var undo = {
        type : ''
    }


    /////////////////////////////////////////////////////////////////////////////
    // Returns true if the file with the given id is on the map.
    self.hasFile = function(id)
    {
        files.forEach(function(f) {
            if (f.id == id) return true;
        });
        return false;
    }

    
    /////////////////////////////////////////////////////////////////////////////
    //
    function removefilePopup(keepState) {
        if (self.filePopup.popup) {
            self.filePopup.popup.remove();
        }
        if (!keepState) {
            self.filePopup.index = 0;
            self.filePopup.files = 0;
            self.filePopup.gps = {}
            self.filePopup.fileId = 0;
            self.filePopup.coordinate = null;
            self.filePopup.feature = null;
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Remove all popups
    function closePopup() {
        removefilePopup(false);
        
        if (self.targetPopup.popup) {
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
        }
        
        if (self.locationPopup.popup) {
            self.locationPopup.popup.remove();
            self.locationPopup.popup = 0;
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Move to the next/previous file in the popup
    function filePopupNext(dir) {

        var cur = self.filePopup.index;

        if (dir == 1) {
            if (self.filePopup.index+1 < self.filePopup.files) {
                self.filePopup.index++;
            }
            else {
                self.filePopup.index = 0;
            }
        }
        else {
            if (self.filePopup.index > 0) {
                self.filePopup.index--;
            }
            else if (self.filePopup.files > 0) {
                self.filePopup.index = self.filePopup.files-1;
            }
        }

        // Always create a new popup, even if there was no change.
        // OL may otherwise close the popup.
        createFilePopup(self.filePopup.coordinate, self.filePopup.feature, true);
    }

    /////////////////////////////////////////////////////////////////////////////
    // Returns the first feature if feature is a cluster, else the feature itself.
    function getPrimaryFeature(feature)
    {
        var features = feature.get('features');
        if (features && features.length > 0) {
            return features[0];
        }
        else {
            return feature;
        }
    }
    

    /////////////////////////////////////////////////////////////////////////////
    // Returns the id for the specified feature. If the feature is a cluster,
    // returns the id of the first element in the cluster.
    function getFeatureId(feature)
    {
        return getPrimaryFeature(feature).get('id');
    }


    /////////////////////////////////////////////////////////////////////////////
    // Returns the type of the specified feature. 
    function getFeatureType(feature)
    {
        return getPrimaryFeature(feature).get('featureType');
    }


    /////////////////////////////////////////////////////////////////////////////
    // This method produces the data for the file popup element
    function createFilePopup(Coordinate, feature, keepState) {
        var bres = false;

        removefilePopup(keepState);

        self.filePopup.feature = feature;

        $("#olmap").append('<div id="olmap-popup"></div>')
        self.filePopup.popup = $('#olmap-popup');
        self.popup.setElement(document.getElementById('olmap-popup'));

        var id = 0;
        var features = feature.get('features');
        if (features && features.length > 0) {

            self.filePopup.files = features.length;

            // Cluster

            if (self.filePopup.index > self.filePopup.files) {
                self.filePopup.index = 0;
            }
            id = features[self.filePopup.index].get('id');
        }
        else {
            // Normal
            id = feature.get('id');
            self.filePopup.index = 0;
            self.filePopup.files = 1;
        }

        if ((id === undefined) || (id == 0)) return;

        self.filePopup.fileId = id;

        // Load data for the popup file from IMWS
        IMWS.get('v1/files',
            {
                id: id,
                fields: 'name,ext,imagewidth,imageheight,rating,label,labelcolor,protected,gps',
                imagesize: 'T',
                tagdesc: "description",
                tagheadline: "headline",
                tagtitle: "title",
                tagcountrycode: "countrycode",
                tagcountry: "country",
                tagstate: "state",
                tagcity: "city",
                taglocation: "location"
                
                // We cannto use gpsposition because that's a Composite tag
                // and it is not filled until the file has been written back.
                // If we change coordinates in the Map panel, this will not be the case.
                // tagalt: "gpsaltitude",
                // tagposition: "gpsposition",
            }).then(function (data) {
                if (data.files.length > 0) {
                    bres = true;
                    var file = data.files[0];

                    // We keep the precise location of this file
                    self.filePopup.gps = file.gps;
                    
                    // NO: We show the decimal coordinates. They are always filled, even if the file has not yet been written back.
                    // var pos = file.lat.replace('deg', '°') + ', ';
                    // pos += file.lng.replace('deg', '°');
                    
                    var pos;
                    
                    if (file.gps.lat && file.gps.lon) {
                        pos = file.gps.lat + ', ' + file.gps.lon;
                    }
                    else if (file.gps.destLat && file.gps.destLon) {
                        pos = file.gps.destLat + ', ' + file.gps.destLon;
                    }
                    else {
                        pos = '';
                    }

                    var content = [
                        '<div id="popup" class="ptc-map-popup container-fluid">',
                        '<img class="col-xs-6 thumb" src="' + (file.protected ? appCore.shieldIconUrl() : appCore.fileUrl(id)) + '" />',
                        '  <div class="row">',
                        '    <div class="fname">' + file.name + '</div>',
                        '    <div class="coord">' + pos + '</div>',
                        '    <div class="data">' + IMatch.safeTagValue(file.desc) + '</div>',
                        '  </div>',

                        // Rating & Label
                        '<div class="colctrl">',
                        '    <span class="rating">{rating}</span>',
                        '    <span ng-show="file.label.length" class="label badge" style="background-color: ' + IMatch.safeTagValue(file.labelColor) + '" title="' + IMatch.safeTagValue(file.label) + '">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</span>',
                        '</div>'
                    ].join('\n');

                    // Set the rating

                    if (file.rating == -1) {
                        // Show the reject icon
                        content = content.replace('{rating}','<i class="ratingitem fa fa-ban fa-lg"></i>');
                    }
                    else if (file.rating > 0) {
                        // Show the stars
                        var s = '';
                        for (var i = 0; i < file.rating; ++i) {
                            s += '<i class="ratingitem fa fa-star fa-lg"></i>';
                        }
                        content = content.replace('{rating}',s);
                    }
                    else {
                        content = content.replace('{rating}','');
                    }

                    if ((file.country != undefined) || (file.city != undefined) || (file.location != undefined)) {
                        var ftr = [
                            '  <div class="row">',
                            '    <div class="footer">',
                            '      <span>' + IMatch.safeTagValue(file.country) + '</span><br/>',
                            '      <span>' + IMatch.safeTagValue(file.city) + '</span><br/>',
                            '      <span>' + IMatch.safeTagValue(file.location) + '</span><br/>',
                            '    </div>',
                            '  </div>'
                        ].join('\n');

                        content = content.concat(ftr);
                    }

                    // If this is a cluster, we add controls to navigate between the different files
                    if (self.filePopup.files > 1) {
                        var nav = [
                            '  <div class="row">',
                            '    <div class="navigator">',
                            '      <a id="popup-prev" href="#"><i class="fa fa-chevron-circle-left fa-2x ptc-m-r-10"></i></a>',
                            '      <a id="popup-next" href="#"><i class="fa fa-chevron-circle-right fa-2x ptc-m-r-10"></i></a>',
                            // n of m files (cluster)
                            '      <span> ' + (self.filePopup.index + 1) + ' / '+ self.filePopup.files + '</span>',
                            '    </div>',
                            '  </div>'
                        ].join('\n');
                        content = content.concat(nav);
                    }

                    // The toolbar at the bottom

                    var btnToolbar = [
                        '<div class="row">',
                        '  <div class="btn-toolbar pull-right">',
                        '    <button id="btn-target-marker" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_TARGET"><img src="img/target.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-create-loc" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_LOC"><img src="img/location.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-create-dest" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_DIRECTION"><img src="img/destination.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-goto" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_GOTO"><i class="fa fa-th-large"></i></button>',
                        '    <button id="btn-find-nearby" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_FINDNEARBY"><i class="fa fa-search"></i></button>',
                        '    <button id="btn-reverse-geocode" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_GEOCODE"><i class="fa fa-globe"></i></button>',
                        '    <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>',
                        '  </div>',
                        '</div>',
                    ].join('\n');

                    content = content.concat(btnToolbar);
                    content = content.concat('\n</div>');

                    // Move the map-popup to the proper coordinate
                    self.popup.setPosition(Coordinate);

                    // Remember
                    self.filePopup.coordinate = Coordinate;

                    var options = {
                        animate: true,
                        container: '#olmap-popup',
                        trigger: 'focus',
                        placement: 'auto top',
                        html: true,
                        title: file.title,
                        content: content
                    };

                    self.filePopup.popup.popover(options);
                    self.filePopup.popup.popover('show');

                    // Translate the contents in the popup after it has been added to the DOM.
                    IMatchTranslate.translateElement('#popup');

                    // Button handlers

                    $('#btn-popup-close').click(function(e) {
                        eatLeftClick = true;
                        removefilePopup(false);
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-target-marker').click(function(e) {
                        eatLeftClick = true;
                        var lonlat = ol.proj.transform(self.filePopup.feature.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
                        setTargetMarker(lonlat[0],lonlat[1]);
                        e.preventDefault();
                        return false;
                    });
                    
                    $('#btn-create-loc').click(function(e) {
                        eatLeftClick = true;
                        var lonlat = ol.proj.transform(self.filePopup.feature.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
                        IMatch.createGeoLocation({
                            lat: lonlat[1],
                            lng: lonlat[0],
                            templatefile: self.filePopup.fileId
                        });
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-create-dest').click(function(e) {
                        createDirection(getPrimaryFeature(self.filePopup.feature));
                        removefilePopup(false);
                        e.preventDefault();
                        return false;
                    });
                    

                    $('#btn-goto').click(function(e) {
                        if (self.filePopup.fileId != 0) {
                            IMatch.gotoFolder(null,self.filePopup.fileId);
                        }
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-find-nearby').click(function(e) {
                        var lonlat = ol.proj.transform(self.filePopup.feature.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
                        IMatch.findNearbyFiles({
                            lat: lonlat[1],
                            lng: lonlat[0],
                            alt: 0                
                        }).then(function(response) {
                            if (response.result == 'ok') {
                                appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
                            }
                        });
                        e.preventDefault();
                        return false;
                    });
                    
                    $('#btn-reverse-geocode').click(function(e) {
                        IMatch.reverseGeoCode({
                            id: self.filePopup.fileId
                        }).then(function(response) {
                        });
                        e.preventDefault();
                        return false;
                    });

                    $('#popup-prev').click(function(e) {
                        filePopupNext(0);
                        e.preventDefault();
                        return false;
                    });

                    $('#popup-next').click(function(e) {
                        filePopupNext(1);
                        e.preventDefault();
                        return false;
                    });

                }
            });

        return bres;
    }    


    /////////////////////////////////////////////////////////////////////////////
    // This method produces the data for the popup element
    function createTargetPopup(Coordinate, feature) {

        closePopup();

        eatLeftClick = true;

        $("#olmap").append('<div id="olmap-popup"></div>')
        self.targetPopup.popup = $('#olmap-popup');
        self.popup.setElement(document.getElementById('olmap-popup'));

        var content = [
            '<div id="target-popup" class="ptc-map-popup container-fluid">',
            '  <h4 imt-translate="MSGBOX_APPLY_TARGET_TITLE"></h4>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg btn-color-target imatch-margin-r-10"></i><a href="#" id="btn-target-apply" data-imt-translate="MAP_POPUP_TARGET_APPLY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><img src="img/location.svg" class="imatch-margin-r-10" style="height:1.25em; width:auto;"><a href="#" id="btn-target-create-loc" data-imt-translate="MAP_POPUP_TARGET_CREATE_LOC"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-search fa-lg imatch-margin-r-10"></i><a href="#" id="btn-target-find-nearby" data-imt-translate="MAP_POPUP_BTN_FINDNEARBY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><img src="img/location-remove.svg" class="imatch-margin-r-10" style="height:1.25em; width:auto;"><a href="#" id="btn-target-remove-marker" data-imt-translate="MAP_POPUP_TARGET_REMOVE_MARKER"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="btn-toolbar pull-right">',
            '      <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>',
            '    </div>',
            '  </div>',
            '</div>'
        ].join('\n');

        // Move the map-popup to the proper coordinate
        self.popup.setPosition(Coordinate);

        var options = {
            animate: true,
            container: '#olmap-popup',
            trigger: 'focus',
            placement: 'auto top',
            html: true,
            content: content
        };

        self.targetPopup.popup.popover(options);
        self.targetPopup.popup.popover('show');

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#target-popup');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            eatLeftClick = true;
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            e.preventDefault();
            return false;
        })

        $('#btn-target-apply').click(function(e) {
            eatLeftClick = true;
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            self.onApplyTargetMarker();
            e.preventDefault();
            return false;
        })
        
        $('#btn-target-create-loc').click(function(e) {
            eatLeftClick = true;
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            var lonlat = ol.proj.transform(markerTarget.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
            IMatch.createGeoLocation({
                lat: lonlat[1],
                lng: lonlat[0],
                templatefile: 0
            });
            e.preventDefault();
            return false;
        })

        $('#btn-target-remove-marker').click(function(e) {
            eatLeftClick = true;
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            if (markerTarget) {
                markerSource.removeFeature(markerTarget);
                markerTarget = null;
                appCore.onTargetMarkerChanged(false);
            }
            e.preventDefault();
            return false;
        })
        

        $('#btn-target-find-nearby').click(function(e) {
            eatLeftClick = true;
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            var lonlat = ol.proj.transform(markerTarget.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
            IMatch.findNearbyFiles({
                lat: lonlat[1],
                lng: lonlat[0],
                alt: 0                
            }).then(function(response) {
                if (response.result == 'ok') {
                    appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
                }
            });
            e.preventDefault();
            return false;
        })

        return true;
    }    


    /////////////////////////////////////////////////////////////////////////////
    // This method produces the data for locations
    function createLocationPopup(Coordinate, feature) {

        closePopup();

        eatLeftClick = true;

        self.locationPopup.feature = feature;

        $("#olmap").append('<div id="olmap-popup"></div>')
        self.locationPopup.popup = $('#olmap-popup');
        self.popup.setElement(document.getElementById('olmap-popup'));

        var content = [
            '<div id="location-popup" class="ptc-map-popup container-fluid">',
            '  <h4>{{loc}}</h4>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-pencil fa-lg imatch-margin-r-10"></i><a href="#" id="btn-loc-edit" data-imt-translate="MAP_POPUP_LOC_EDIT"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg btn-color-target imatch-margin-r-10"></i><a href="#" id="btn-loc-apply" data-imt-translate="MAP_POPUP_LOC_APPLY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-search fa-lg imatch-margin-r-10"></i><a href="#" id="btn-loc-find-nearby" data-imt-translate="MAP_POPUP_BTN_FINDNEARBY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <br/><div class="col-sm-12 ptc-popup-row"><i class="fa fa-times-circle fa-lg btn-color-delete imatch-margin-r-10"></i><a href="#" id="btn-loc-delete" data-imt-translate="MAP_POPUP_LOC_DELETE"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="btn-toolbar pull-right">',
            '      <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>',
            '    </div>',
            '  </div>',
            '</div>'
        ].join('\n');

        content = content.replace('{{loc}}',feature.get('name'));

        // Move the map-popup to the proper coordinate
        self.popup.setPosition(Coordinate);

        var options = {
            animate: true,
            container: '#olmap-popup',
            trigger: 'focus',
            placement: 'auto top',
            html: true,
            content: content
        };

        self.locationPopup.popup.popover(options);
        self.locationPopup.popup.popover('show');

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#location-popup');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            self.locationPopup.popup.remove();
            e.preventDefault();
            return false;
        });

        $('#btn-loc-edit').click(function(e) {
            self.locationPopup.popup.remove();

            IMatch.editGeoLocation({
                guid: self.locationPopup.feature.get('guid')
            });

            e.preventDefault();
            return false;
        });
        
        $('#btn-loc-apply').click(function(e) {
            eatLeftClick = true;
            self.locationPopup.popup.remove();
            self.onApplyLocation(self.locationPopup.feature.get('guid'));
            e.preventDefault();
            return false;
        });

        $('#btn-loc-find-nearby').click(function(e) {
            eatLeftClick = true;
            self.locationPopup.popup.remove();
            self.onFindNearbyLocation(self.locationPopup.feature.get('guid'));
            e.preventDefault();
            return false;
        });

        $('#btn-loc-delete').click(function(e) {
            eatLeftClick = true;
            self.locationPopup.popup.remove();
            // Ask the user
            var m = new IMatchModals();
            m.showMessageBoxYesNo({
                headline: IMatchTranslate.getStr('MSGBOX_DELETE_LOC_TITLE'),
                content: IMatchTranslate.getStr('MSGBOX_DELETE_LOC_TEXT'),
                okCallback: function() {
                    clearUndo();
                    IMWS.delete('v1/geolocation',{
                        guid: self.locationPopup.feature.get('guid')
                    });
                }
            });

            e.preventDefault();
            return false;
        });

        return true;
    }    


    /////////////////////////////////////////////////////////////////////////////
    // This method produces the data for the destination point popup element
    function createDestPopup(Coordinate, feature) {

        closePopup();

        eatLeftClick = true;

        $("#olmap").append('<div id="olmap-popup"></div>')
        self.targetPopup.popup = $('#olmap-popup');
        self.popup.setElement(document.getElementById('olmap-popup'));

        var content =
            `<div id="target-popup" class="ptc-map-popup container-fluid">
              <h4 imt-translate="MAP_POPUP_DEST_MARKER"></h4>
              <div class="row">
                <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg btn-color-target imatch-margin-r-10"></i><a href="#" id="btn-dest-apply" data-imt-translate="MAP_POPUP_BTN_DEST_APPY"></a></div>
              </div>
              <div class="row">
                <div class="btn-toolbar pull-right">
                  <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>
                </div>
              </div>
            </div>`;

        // Move the map-popup to the proper coordinate
        self.popup.setPosition(Coordinate);

        var options = {
            animate: true,
            container: '#olmap-popup',
            trigger: 'focus',
            placement: 'auto top',
            html: true,
            content: content
        };

        self.targetPopup.popup.popover(options);
        self.targetPopup.popup.popover('show');

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#target-popup');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            e.preventDefault();
            return false;
        })

        $('#btn-dest-apply').click(function(e) {
            self.targetPopup.popup.remove();
            self.targetPopup.popup = null;
            var lonlat = ol.proj.transform(feature.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');

            IMatch.setGPSCoordinates({
                idlist: IMatch.idlist.fileWindowSelection,
                target: 'dest',
                destlat: lonlat[1],
                destlng: lonlat[0]
            }).then(response => {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });
            
            return false;
        })

        return true;
    }        

    /////////////////////////////////////////////////////////////////////////////
    //
    this.loadMap = function(keepCenter)
    {
        clearMap();

        var keepCenter = keepCenter || false;

        var center = null;
        if (keepCenter && self.olMap && self.olMap.getView()) {
            center = self.olMap.getView().getCenter();
        }
        
        self.olMap = null;

        createMap();

        loadLocations();

        if (keepCenter && (center != null)) {
            self.olMap.getView().setCenter(center);
        }

        self.loadFiles(false);

        self.onTrackLoaded(false);
    }

    // Bing uses a rather obscure language mapping system
    // for map tiles. For example, tiles in German are only available
    // when using culture ='de-de' but not for 'de.
    // We here map 'de' to 'de-de' and similar.
    // See https://docs.microsoft.com/en-us/bingmaps/rest-services/common-parameters-and-types/supported-culture-codes
    this.mapLangBing = function(l)
    {
        const MT = {
            'de': 'de-de',
            'en': 'en-US',
            'es': 'es-ES',
            'fr': 'fr-FR',
            'it': 'it-it',
            'nl': 'nl-BE',
        };

        let res = MT[l];
        return res ? res : l;
    }


    // HERE uses TLC codes ('ger' instead of 'de')
    this.mapLangHERE = function(l)
    {
        const MT = {
            'ar':'ara',
            'ba':'baq',
            'ca':'cat',
            'zh':'chi',
            'cht':'cht',
            'cs':'cze',
            'da':'dan',
            'nl':'dut',
            'en':'eng',
            'fi':'fin',
            'fr':'fre',
            'de':'ger',
            'ga':'gle',
            'el':'gre',
            'he':'heb',
            'hi':'hin',
            'id':'ind',
            'it':'ita',
            'mul':'mul',
            'no':'nor',
            'fa':'per',
            'pl':'pol',
            'pt':'por',
            'ru':'rus',
            'si':'sin',
            'es':'spa',
            'sv':'swe',
            'th':'tha',
            'tr':'tur',
            'uk':'ukr',
            'ur':'urd',
            'vi':'vie',
            'cy':'wel',
        }

        let res = MT[l];
        return res ? res : l;
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Create the map, set everyting up.
    function createMap()
    {
        clearStyleCache();

        vectorSource = new ol.source.Vector({});
        markerSource = new ol.source.Vector({});
        locationSource = new ol.source.Vector({});
        trackSource = new ol.source.Vector({});
        trackMarkerSource = new ol.source.Vector({});
        lineSource = new ol.source.Vector({});
        
        fileRouteSource = new ol.source.Vector({});

        clusterSource = new ol.source.Cluster({
            distance: 5,  // Match to markerclusterer in gmaps
            threshold: 2,
            source: vectorSource
            // 2021: We add the destination markers (loc shown) also
            // to the vectorSource. This folds dest markers into clusters,
            // just like file markers. Not sure if this will be problem,
            // but IMatch 2020 did never show the target markers.
            /*,
            geometryFunction: function(feature) {
                // We ignore markers of type 'dest'
                if (getFeatureType(feature) == 'dest') {
                    return null;
                }
                else {
                    return feature.getGeometry();
                }
            }*/
        });

        // This layer displays the file markers / thumbnails
        fileLayer = new ol.layer.Vector(
            {
                source: clusterSource,
                style: fileStyle,
                zindex: 10
            });

        // This layer is used for the target marker etc.
        markerLayer = new ol.layer.Vector(
            {
                source: markerSource,
                style: markerStyle
            });

        // This layer is used for the lines between source and destination markers
        lineLayer = new ol.layer.Vector(
            {
                source: lineSource,
                style: lineStyle
            });
                   

        locationLayer = new ol.layer.Vector(
            {
                source: locationSource,
                style: locationStyle
            });

        trackLayer = new ol.layer.Vector(
            {
                source: trackSource,
                style: trackStyle
            });

        trackMarkerLayer = new ol.layer.Vector(
            {
                source: trackMarkerSource,
                style: trackMarkerStyle
            });

        trackMarkerDataLayer = new ol.layer.Vector(
            {
                source: trackMarkerSource,
                style: trackMarkerDataStyle
            });

        // This layer shows data like file names etc.
        dataLayer = new ol.layer.Vector(
            {
                source: clusterSource,
                style: dataStyle
            });

        dataLayer.setVisible(appCore.appConfig.showFileData);

        fileRouteLayer = new ol.layer.Vector(
            {
                source: fileRouteSource,
                style: fileRouteStyle
            });

        // The actual map

        self.olMap = new ol.Map({
            target: 'olmap',  // the div to link to

            controls: ol.control.defaults({
                zoom: true,
                attribution: true,
                rotate: true
            }),

            interaction: ol.interaction.defaults(),

            view: new ol.View({
                center: ol.proj.fromLonLat([appCore.appConfig.mapCenter.lng,appCore.appConfig.mapCenter.lat]),
                
                zoom: Math.min(21, Math.max(1, appCore.appConfig.zoomLevel)),

                // Limit the maximum zoom to the maximum supported by the map type
                maxZoom: appCore.getActiveMapConfig().maxZoom
            })
        });

        // Set the source based on the current map type

        switch (appCore.getActiveMapConfig().type) {
            case 'osm':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.OSM({ layer: 'osm' })
                }));
                break;

            case 'stamen':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.Stamen({
                        layer: 'watercolor',
                        attributions: 'Map and tiles by <a href="http://stamen.com">Stamen Design</a>, under <a href="http://creativecommons.org/licenses/by/3.0">CC BY 3.0</a>. Data by <a href="http://openstreetmap.org">OpenStreetMap</a>, under <a href="http://creativecommons.org/licenses/by-sa/3.0">CC BY SA</a>.'
                    })
                }));
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.Stamen({ layer: 'terrain-labels' })
                }));
                break;

            case 'cycleosm':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.OSM ({
                        url: 'https://{a-c}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
                        attributions: 'Map and tiles by <a href="https://github.com/cyclosm/cyclosm-cartocss-style/releases" title="CyclOSM - Open Bicycle render">CyclOSM</a>',
                    }),
                }));
                break;

                case 'opentopomap':
                    self.olMap.addLayer(new ol.layer.Tile({
                        source: new ol.source.OSM ({
                            url: 'https://{a-c}.tile.opentopomap.org/{z}/{x}/{y}.png',
                            attributions: 'Kartendaten: © OpenStreetMap-Mitwirkende, SRTM | Kartendarstellung: © OpenTopoMap (CC-BY-SA)',
                        }),
                    }));
                    break;
   
            // Bing

            case 'bing-aerial':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.BingMaps({ key: appCore.IMatchConfig.bingMapsKey, imagerySet: 'Aerial', culture: self.mapLangBing(appCore.IMatchConfig.locale) })
                }));
                break;

            case 'bing-aeriallbl':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.BingMaps({ key: appCore.IMatchConfig.bingMapsKey, imagerySet: 'AerialWithLabelsOnDemand', culture: self.mapLangBing(appCore.IMatchConfig.locale) })
                }));
                break;

            case 'bing-road':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.BingMaps({ key: appCore.IMatchConfig.bingMapsKey, imagerySet: 'Road', culture: self.mapLangBing(appCore.IMatchConfig.locale) })
                }));
                break;

            case 'bing-ordnanceSurvey':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.BingMaps({ key: appCore.IMatchConfig.bingMapsKey, imagerySet: 'ordnanceSurvey', culture: self.mapLangBing(appCore.IMatchConfig.locale) })
                }));
                break;

            case 'bing-birdseye':
                self.olMap.addLayer(new ol.layer.Tile({
                    source: new ol.source.BingMaps({ key: appCore.IMatchConfig.bingMapsKey, imagerySet: 'BirdseyeV2WithLabels', culture: self.mapLangBing(appCore.IMatchConfig.locale) })
                }));
                break;
                    
            // HERE

            case 'here-normal-day':
            case 'here-normal-night':
            case 'here-pedestrian-day':
            case 'here-pedestrian-night':
            case 'here-terrain-day':
            case 'here-satellite-day':
            case 'here-hybrid-day':
                {
                    let ld = appCore.getActiveMapConfig().layerData;
                    let key = appCore.IMatchConfig.HEREMapsKey.split(':');
                    if (key.length >= 2) {
                        self.olMap.addLayer(new ol.layer.Tile({
                            source: new ol.source.XYZ({
                                preload: Infinity,
                                url: `https://{1-4}.${ld.base}.maps.api.here.com/${ld.type}/2.1/maptile/newest/${ld.scheme}/{z}/{x}/{y}/256/png?app_id=${key[0]}&app_code=${key[1]}&lg=${self.mapLangHERE(appCore.IMatchConfig.locale)}`,
                                attributions: `Map Tiles &copy; ${new Date().getFullYear()} <a style="color:#000;" href="https://www.here.com">HERE</a>`
                            })

                        }));
                    }
                    else {
                        // We assume an API key (no : in the key)
                        self.olMap.addLayer(new ol.layer.Tile({
                            source: new ol.source.XYZ({
                                preload: Infinity,
                                url: `https://{1-4}.${ld.base}.maps.ls.hereapi.com/${ld.type}/2.1/maptile/newest/${ld.scheme}/{z}/{x}/{y}/256/png?apiKey=${appCore.IMatchConfig.HEREMapsKey}&lg=${self.mapLangHERE(appCore.IMatchConfig.locale)}`,
                                attributions: `Map Tiles &copy; ${new Date().getFullYear()} <a style="color:#000;" href="https://www.here.com">HERE</a>`
                            })

                        }));
                    }
                }
                break;
        }

        self.olMap.addLayer(fileLayer);
        self.olMap.addLayer(markerLayer);
        self.olMap.addLayer(dataLayer);

        self.olMap.addLayer(lineLayer);

        self.olMap.addLayer(locationLayer);
        locationLayer.setVisible(appCore.appConfig.showLocations);

        self.olMap.addLayer(trackLayer);
        self.olMap.addLayer(trackMarkerLayer);
        self.olMap.addLayer(trackMarkerDataLayer);

        self.olMap.addLayer(fileRouteLayer);
        
        // Add controls

        var ctrlScaleLine = new ol.control.ScaleLine({
            units: 'metric',  // Supported values are 'degrees', 'imperial', 'nautical', 'metric', 'us'.
            minWidth: 100
        });
        self.olMap.addControl(ctrlScaleLine);


        var ctrlFullScreen = new ol.control.FullScreen({
            tipLabel: IMatchTranslate.getStr('T_FULL_SCREEN_MODE')
        });
        self.olMap.addControl(ctrlFullScreen);

        self.olMap.addControl(new ol.control.ZoomSlider());

        //var ctrlOverview = new ol.control.OverviewMap({
        //    collapsible : true,
        //    tipLabel: gettextCatalog.getString('@tl-fw-map-btn-overview')
        //});
        //self.olMap.addControl(ctrlOverview);


        //var ctrlZoom2Ext = new ol.control.ZoomToExtent({
        //    collapsible: true,
        //    tipLabel: gettextCatalog.getString('@tl-fw-map-btn-zoom2ext')
        //});
        //self.olMap.addControl(ctrlZoom2Ext);

        
        // Add the popup we use when the user clicks a file feature

        self.popup = new ol.Overlay({
            //element: document.getElementById('olmap-popup'),
            positioning: 'top-center',
            autoPan: true,
            autoPanAnimation: {
                duration: 250
            }
        });

        self.olMap.addOverlay(self.popup);

        // Handle single-clicks on the map

        self.olMap.on('singleclick', function (e) {
            // Find the feature the user has clicked.
            if (eatLeftClick) {
                eatLeftClick = false;
                return;
            }
            var features = [];
            var feature = self.olMap.forEachFeatureAtPixel(e.pixel,
                    function (feature, layer) {
                        var type = getFeatureType(feature);
                        if (type != undefined) {
                            return feature;
                        }
                    },20); // 20 pixel tolerance

            // If a feature was found, open the popup
            if (feature) {
                closePopup();
                
                var type = getFeatureType(feature);
                if (type == 'file') {
                    //createFilePopup(e.coordinate, feature, false);
                    setFocusedMarker(feature,true);
                }
                else if (type == 'target') {
                    //createTargetPopup(e.coordinate, feature, false);
                }
                else if (type == 'location') {
                    //createLocationPopup(e.coordinate, feature, false);
                }
                else if (type == 'dest') {
                    // If we click a dest marker, we make the corresponding 
                    // source marker the active focused marker.
                    let f = getPrimaryFeature(feature);
                    if (f && f.get('parentFeature')) {
                        setFocusedMarker(f.get('parentFeature'),true);
                    }
                }
            }
            else {
                closePopup();
                var lonlat = ol.proj.transform(e.coordinate, 'EPSG:3857', 'EPSG:4326');
                setTargetMarker(lonlat[0],lonlat[1]);
            }

            e.preventDefault();
            return false;

        });  // singleclick

        // Handle right-clicks
        
        self.olMap.getViewport().addEventListener('contextmenu', function (e) {

            // Find the feature the user has clicked.
            var coord = self.olMap.getEventCoordinate(e);
            var features = [];
            var feature = self.olMap.forEachFeatureAtPixel([e.offsetX,e.offsetY],
                    function (feature, layer) {
                        var type = getFeatureType(feature);
                        if (type != undefined) {
                            return feature;
                        }
                    },20); // 20 pixel tolerance

            // If a feature was found, open the popup
            if (feature) {
                closePopup();
                
                var type = getFeatureType(feature);
                if (type == 'file') {
                    setFocusedMarker(feature,true);
                    createFilePopup(coord, feature, false);
                }
                else if (type == 'target') {
                    createTargetPopup(coord, feature, false);
                }
                else if (type == 'location') {
                    createLocationPopup(coord, feature, false);
                }
                else if (type == 'dest') {
                    let f = getPrimaryFeature(feature);
                    if (f && f.get('parentFeature')) {
                        setFocusedMarker(f.get('parentFeature'),true);
                    }
                    createDestPopup(coord, feature, false);
                }

                e.preventDefault();
                return false;
            }

            e.preventDefault();
            return false;
        })
        

        // We watch for zoom change so we can update the user settings.

        self.olMap.getView().on('change:resolution', IMatch.debounce(function (evt) {
            saveConfig();
            // We need to clear all styles for cluster nodes (the text displayed
            // needs to change if the zoom changes and the cluster updates).
            clearStyleCache();

            window.setTimeout(function() {
                createFOV();
            },500);

        },50));

        // We watch for zoom change so we can update the user settings.

        self.olMap.getView().on('change:center', IMatch.debounce(function (evt) {
            saveConfig();
        },1000));
    }

    /////////////////////////////////////////////////////////////////////////////
    // Loads all locations from IMWS and creates features for them on the location layer.
    function loadLocations()
    {
        if (diLocations) {
            self.olMap.removeInteraction(diLocations);
        }
        
        locationSource.clear();

        locations = [];

        // We use the change event to keep track of the last modified feature.
        // OL is not very clever and does not tell us in modifyend which feature was modified... :-/
        diLocationsLMF = null;

        IMWS.get('v1/geolocation').then(function(response) {
            locations = response.locations;
            var features = [];
            locations.forEach(function(l) {
                var options = {
                    // Our custom data
                    id : l.guid,
                    featureType: 'location',
                    guid: l.guid,
                    name: l.name,
                    imatchLocation: l,
                    mapRadius: l.mapRadius,
                    mapType: l.mapType,
                    // geometry: new ol.geom.Point(ol.proj.fromLonLat([l.lng,l.lat]))
                    geometry: new ol.geom.Circle(ol.proj.fromLonLat([l.lng,l.lat]),l.mapRadius)
                }
                var loc = new ol.Feature(options);        

                // Track changes. And then we use this in modifyend below
                loc.on('change',function(e) {
                    diLocationsLMF = e.target;
                });

                features.push(loc);
            }, this);

            if (features.length > 0) {

                locationSource.addFeatures(features);

                var ft = new ol.Collection(features);

                diLocations = new ol.interaction.Modify({
                    features: ft,
                    pixelTolerance: 20
                });

                // Add the event to the drag and drop feature
                diLocations.on('modifyend',function(e) {
                    if (diLocationsLMF == null) {
                    }
                    else {
                        undo.type = 'location';
                        undo.location = diLocationsLMF.get('imatchLocation');
                        appCore.setUndo(true);

                        var lonlat = ol.proj.transform(diLocationsLMF.getGeometry().getCenter(), 'EPSG:3857', 'EPSG:4326');

                        var center = diLocationsLMF.getGeometry().getCenter();
                        var radius = diLocationsLMF.getGeometry().getRadius();

                        IMWS.post('v1/geolocation/update',{
                            guid: diLocationsLMF.get('guid'),
                            lat: lonlat[1],
                            lng: lonlat[0],
                            radius: radius
                        }).then(function(response) {
                            if (response.result != 'ok') {
                            }
                        });
                    }
                },diLocations);

                if (!appCore.appConfig.lockMove) {
                    self.olMap.addInteraction(diLocations);
                }
            }
        });
    }

    /////////////////////////////////////////////////////////////////////////////
    // Each file marker has it's own style, managed by the file id.
    // This allows us per-file styling and images
    function fileStyle(feature,resolution) {

        var id, h, w, name, prot;

        // This may be a cluster of files or a single file
        var features = feature.get('features');
        var f = null;
        var files = 1;
        if (features) {
            files = features.length;
            if (features.length > 0) {
                // For clusters, we use the first file
                f = features[0];
            }
        }
        else {
            // Not a cluster
            f = feature;
        }

        // This should not happen else something is really wrong
        if (f == null) return null;

        id = f.get('id');

        if (!styleCache[id]) {
            let focused = f.get('focused');
            let type = f.get('featureType');

            prot = f.get('protected') === true;
            w = f.get('imageWidth');
            h = f.get('imageHeight');

            // Do we show images?
            if (appCore.appConfig.showImages && (type == 'file')) {
                if (prot) {
                    w = h = 32;
                }
                var s = Math.min(1.0, 80.0 / Math.max(1, w));
                styleCache[id] = new ol.style.Style({
                    image: new ol.style.Icon({
                        scale: s,
                        // Use the shield icon or the thumbnail
                        src: prot ? appCore.shieldIconUrl() : appCore.fileUrl(id)
                    })
                });
            }
            else {
                let img = null;
                let txt = null;

                if (files > 1) {
                    // We emulate the way the markerclusterer works for the Google Maps
                    // by using the same icons and colors, depending on the number of files.

                    img = new ol.style.Icon(({
                        src: 'gm/gmaps-clusterer/images/m' + calcFileStyle(files,5) + '.png'
                    }));

                    txt = new ol.style.Text({
                        text: '' + files,
                        font: 'bold 12pt "Helvetica Neue",Helvetica,Arial,sans-serif ',
                        stroke: new ol.style.Stroke({
                            width: 1,
                            color: 'rgba(255,255,255,0.75)'
                        }),
                        fill: new ol.style.Fill({
                            color: 'rgba(0,0,0,1)',
                        })
                    });
                }
                else {
                    // We use a FA icon
                    if (type == 'file') {
                        // This is just for easier hit testing. It's below the marker icon and invisible.
                        img = new ol.style.Circle({
                            radius: 30,
                            snapToPixel: false,
                            fill: new ol.style.Fill({
                                color: 'rgba(255,255,0,0.01)'  // Almost invisible but still there and clickable
                            })
                        });
                        
                        txt = new ol.style.Text({
                            text: '\uf041',
                            font: 'normal 24pt FontAwesome',
                            textBaseline: 'bottom',
                            stroke: new ol.style.Stroke({
                                width: files > 1 ? 4 : 2,
                                color: focused ? 'rgba(255,255,0,0.75)' : 'rgba(255,255,255,0.75)'
                            }),
                            fill: new ol.style.Fill({
                                color: focused ? 'rgba(215,41,100,0.90)' : 'rgba(192,0,192,0.75)',
                            })
                        });
                    }
                    else {
                        img = new ol.style.Icon(({
                            src: 'img/destination.png',
                            snapToPixel : false
                        }));
                    }
                }

                styleCache[id] = new ol.style.Style({
                    image: img,
                    text: txt,
                });
            }
        }

        return [styleCache[id]];
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // see : MarkerClusterer.prototype.calculator_ = function(markers, numStyles) in 
    // ...mapapp\gm\gmaps-clusterer\markerclusterer.js
    function calcFileStyle(fileCount, numStyles) {
        var index = 0;
        var count = fileCount;
        var dv = count;
        while (dv !== 0) {
            dv = parseInt(dv / 10, 10);
            index++;
        }
        index = Math.min(index, numStyles);
        return index;
    }

    // Scale the font-size
    function calcFontSize(baseSize,resolution)
    {
        var res = baseSize;
        if (resolution >= 10) {
            res = 0;
        }
        else if (resolution >= 5) {
            res = (baseSize / 3.0) * 2.0;
        }
        return res;
    }


    /////////////////////////////////////////////////////////////////////////////
    // Get the data style for each file
    function dataStyle(feature, resolution) {

        var id, h, w, name, prot;

        // This may be a cluster of files or a single file
        let f = getPrimaryFeature(feature);
        if (f == null) return null;
        if (f.get('featureType') != 'file') return;

        var features = feature.get('features');

        id = 'd' + f.get('id');
        prot = f.get('protected') === true;
        w = f.get('imageWidth');
        h = f.get('imageHeight');

        if (features && (features.length > 1)) {
            // Indicate that there are multiple files
            if (features.length > 2) {
                name = IMatchTranslate.getStr('MAP_FILE_COUNT')
            }
            else {
                name = IMatchTranslate.getStr('MAP_FILE_COUNT_ONE')
            }
            name = name.replace('{filename}', f.get('name'));
            name = name.replace('{0}', features.length - 1);
        }
        else {
            // Use the file name
            name = f.get('name');
        }

        if (!styleCache[id]) {

            var tofsy = 20;

            // If we show images, we need to offset the text by the image height
            if (appCore.appConfig.showImages) {
                // Protected files use a 32x32 icon
                if (prot) {
                    w = h = 32;
                }
                var s = Math.min(1.0, 80.0 / Math.max(1, w));
                tofsy += (s * h / 2);
            }

            // We choose the text and stroke color for the labels based on the map type
            var fillColor = '#000000';
            var strokeColor = '#ffffff';
            var strokeWidth = 1;

            switch (appCore.getActiveMapConfig().type) {
                case 'bing-aerial':
                case 'bing-aeriallbl':
                    fillColor = '#ffffff';
                    strokeColor = '#ffffff';
                    strokeWidth = 0;
                    break;
            }

            styleCache[id] = new ol.style.Style({
                text: new ol.style.Text({
                    text: name,
                    font: calcFontSize(12,resolution) + 'pt "Helvetica Neue",Helvetica,Arial,sans-serif ',
                    offsetY: tofsy,
                    fill: new ol.style.Fill({
                        color: fillColor
                    }),
                    stroke: new ol.style.Stroke({
                        width: strokeWidth,
                        color: strokeColor
                    })
                })
            });

        }

        return [styleCache[id]];
    }


    /////////////////////////////////////////////////////////////////////////////
    // Styles for markers
    function markerStyle(feature, resolution) {

        var f = feature;

        var id = f.get('id');

        var style = new ol.style.Style({
            // We use a FA icon
            text : new ol.style.Text({
                text: '\uf041',
                //text: '\uf276',
                font: 'normal 32pt FontAwesome',
                textBaseline: 'bottom',
                stroke: new ol.style.Stroke({
                    width: 4,
                    color: 'rgba(0,0,0,0.75)'
                }),
                fill: new ol.style.Fill({
                    color: 'rgba(255,200,32,0.90)',
                })
            })
        });

        return style;
    }


    /////////////////////////////////////////////////////////////////////////////
    // Style for locations
    function locationStyle(feature, resolution)
    {
        var style = new ol.style.Style({

            //geometry: new ol.geom.Circle(feature.getGeometry().getCoordinates(), feature.get('mapRadius')),
            
            stroke: new ol.style.Stroke({
                width: 4,
                color: 'rgba(255,0,0,0.75)'
            }),

            fill: new ol.style.Fill({
                color: 'rgba(255,255,0,0.5)',
            }),

            text : new ol.style.Text({
                text: feature.get('name'),
                font: 'bold ' + calcFontSize(12,resolution) + 'pt "Helvetica Neue",Helvetica,Arial,sans-serif ',
                textBaseline: 'middle',
                stroke: new ol.style.Stroke({
                    width: 1,
                    color: 'rgba(255,255,255,0.75)'
                }),
                fill: new ol.style.Fill({
                    color: 'rgba(0,0,0,1)',
                })
            })
        });
        
        return style;
    }


    /////////////////////////////////////////////////////////////////////////////
    // Style for tracks
    function trackStyle(feature, resolution)
    {
        if (feature.style != null) {
            return feature.style;
        }
        else {
            var style = new ol.style.Style({
                stroke: new ol.style.Stroke({
                    width: 6,
                    color: 'rgba(0,0,255,0.5)'
                })
            });
            
            return style;
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Style for file route
    function fileRouteStyle(feature, resolution)
    {
        var geometry = feature.getGeometry();

        var styles = [
            // linestring
            new ol.style.Style({
              stroke: new ol.style.Stroke({
                width: 6,
                color: 'rgba(70,0,255,0.5)'
              }),
            })
        ];

        var last;
        geometry.forEachSegment(function (start, end) {
            var dx = end[0] - start[0];
            var dy = end[1] - start[1];
            var rotation = Math.atan2(dy, dx);

            // We use the center point between end and last
            // for the arrow (middle of the line segment).
            var pt;
            if (last) {
                pt = [
                    (end[0] + last[0])/2,
                    (end[1] + last[1])/2
                ];
            }
            else {
                pt = end;
            }
            // DEEP copy!
            last = [end[0],end[1]];
                        
            // arrows
            styles.push(
              new ol.style.Style({
                geometry: new ol.geom.Point(pt),
                image: new ol.style.Icon({
                    // NOTE: The SVG must have a width and height in px.
                  src: 'img/chevron.svg',
                  anchor: [0.5, 0.5],
                  scale: 0.075 / Math.pow(resolution, 1 / 2),
                  rotateWithView: true,
                  rotation: -rotation,
                }),
              })
            );
          });
          
        return styles;
    }

    /////////////////////////////////////////////////////////////////////////////
    // Style for track markers
    function trackMarkerStyle(feature, resolution)
    {
        var f = feature;

        var id = f.get('id');

        var style = new ol.style.Style({
           
            // We use a FA icon
            text : new ol.style.Text({
                text: '\uf041',
                //text: '\uf276',
                font: 'normal 24pt FontAwesome',
                textBaseline: 'bottom',
                stroke: new ol.style.Stroke({
                    width: 4,
                    color: 'rgba(200,200,255,0.75)'
                }),
                fill: new ol.style.Fill({
                    color: 'rgba(0,0,255,1)',
                })
            })
        });

        return style;
    }

    /////////////////////////////////////////////////////////////////////////////
    // Style for the data (file name) for track markers
    function trackMarkerDataStyle(feature, resolution)
    {
        var f = feature;

        var id = f.get('id');

        var style = new ol.style.Style({

            text: new ol.style.Text({
                text: appCore.appConfig.showFileData ? f.get('name') : null,
                font: calcFontSize(12,resolution) + 'pt "Helvetica Neue",Helvetica,Arial,sans-serif ',
                offsetY: 20,
                fill: new ol.style.Fill({
                    color: '#000000'
                }),
                stroke: new ol.style.Stroke({
                    width: 1,
                    color: '#ffffff'
                })
            })
           
        });

        return style;
    }

    /////////////////////////////////////////////////////////////////////////////
    // For lines between source and destination marker.
    function lineStyle(feature, resolution)
    {
        return new ol.style.Style({
            fill: new ol.style.Fill({
                weight: 1,
                color: 'rgba(255,255,0.6)'
            }),
            stroke: new ol.style.Stroke({
                width: 4,
                color: 'rgba(200,0,0,1)'
            })
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    // Clear the Undo
    function clearUndo()
    {
        undo = {
            type : ''
        }
        appCore.setUndo(false);
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function clearMap()
    {
        closePopup();

        removeSelectTool();

        setFocusedMarker(null);

        if (fovLayer.source != null) {
            fovLayer.source.clear(true);
        }
        
        if (diFiles) {
            self.olMap.removeInteraction(diFiles);
            diFiles = null;
            diFilesLMF = null;
        }

        if (vectorSource) {
            vectorSource.clear(true);
        }

        if (clusterSource) {
            clusterSource.clear();
        }

        if (lineSource) {
            lineSource.clear(true);
        }

        clearStyleCache();
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function clearStyleCache() {
        styleCache = {};
    }

    
    var loadFilesDelayed = IMatch.debounce(function (keepCenter) {
        self.loadFiles(keepCenter || false);
    },2000);


    /////////////////////////////////////////////////////////////////////////////
    // Load the files in the file window
    self.loadFiles = function(keepCenter)
    {
        clearMap();

        files = [];

        var params = {
            idlist: IMatch.idlist.fileWindowSelection,
            fields: 'id,folderid,gps,name,datetime',
            rtagfocalLength: 'focallength',
            rtagfocallength35mm: 'focallength35mm',
            tagmake: 'make',
            tagmodel: 'model',
            sortby: 'datetime-Asc'
        }

        if (appCore.appConfig.showImages) {
            params.fields += ',imagewidth,imageheight,protected';
            params.imagesize = 'T';
        }

        var keepCenter = keepCenter || false;
        var focus = 0;
        var center = keepCenter ? self.olMap.getView().getCenter() : null;
        var features = [];

        // Get the focused file
        IMWS.get('v1/idlist',{
            name: IMatch.idlist.fileWindowFocusedFile
        }).then(function(response) {
            if (response.id.length > 0) {
                focus = response.id[0]
            }
            IMWS.get('v1/files',params).then(function(response) {

                // We keep a copy
                files = response.files;

                if (appCore.debugMode()) {
                    console.log('OL loaded ' + response.files.length + ' files.');
                }

                files.forEach(function(f) {

                    let fileCoords = MapUtility.getSafeCoordinates(f);

                    // Skip files without GPS coordinates
                    if (fileCoords !== undefined) {
                        // Remember the first file coordinates
                        if (!keepCenter && (center == null)) {
                            center = fileCoords;
                        }

                        // Create the feature
                        var options = {
                            id: f.id,
                            gps: f.gps,
                            featureType: 'file',
                            focused: f.id == focus,
                            name: f.name,
                            coordType: fileCoords.type,
                            file: f,
                            geometry: new ol.geom.Point(ol.proj.fromLonLat([fileCoords.lon, fileCoords.lat])),
                        };

                        if (appCore.appConfig.showImages) {
                            options.image = f.image;

                            // In this case no image was available on the server. Since we now load the default image in the fileUrl call,
                            // we set the default dimensions
                            if ((f.imageWidth == 0) || (f.imageHeight == 0)) {
                                options.imageWidth = 256;
                                options.imageHeight = 256;
                            }
                            else {
                                options.imageWidth = f.imageWidth;
                                options.imageHeight = f.imageHeight;
                            }
                            options.protected = f.protected;
                        }

                        var marker = new ol.Feature(options);
                        features.push(marker);

                        // The intiially focused marker
                        if (options.focused) {
                            setFocusedMarker(marker);
                        }

                        // Keep track of the last changed feature
                        marker.on('change',function(e) {
                            diFilesLMF = e.target;
                        });

                        // Does the file have a dest GPS record?
                        // If so, add the destination marker.
                        if (f.gps.destLat && f.gps.destLon) {
                            createDestinationMarker(marker);
                        }
                    }
                }, this);

                if (features.length > 0) {

                    vectorSource.addFeatures(features);
                    if (!keepCenter && (center != null) && appCore.appConfig.followFocus) {
                        window.setTimeout(function () {
                            self.olMap.getView().setCenter(ol.proj.fromLonLat([center.lon, center.lat]));
                        }, 50);
                    }

                    if (diFiles) {
                        self.olMap.removeInteraction(diFiles);
                        diFilesLMF = null;
                    }


                    // Add the event to the drag and drop feature
                    diFiles = new ol.interaction.Modify({
                        //features: new ol.Collection(features),
                        source: vectorSource,
                        style: null,
                        pixelTolerance: 30
                    });

                    if (!appCore.dbReadOnly && !appCore.appConfig.lockMove) {
                        self.olMap.addInteraction(diFiles);
                    }

                    diFiles.on('modifyend',function(e) {
                        if (diFilesLMF == null) return;

                        undo.type = diFilesLMF.get('featureType');
                        undo.id = diFilesLMF.get('id');
                        undo.gps = diFilesLMF.get('gps');

                        // Update the file in the database after the user has dragged the marker
                        if (undo.type == 'file') {
                            let lonlat = ol.proj.transform(diFilesLMF.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');

                            let params = {
                                target: diFilesLMF.get('coordType'),
                                id: getFeatureId(diFilesLMF)
                            }
                            if (diFilesLMF.get('coordType') == 'created') {
                                params.lng = lonlat[0],
                                params.lat = lonlat[1]
                            }
                            else {
                                params.destlng = lonlat[0],
                                params.destlat = lonlat[1]
                            }
                            
                            IMatch.setGPSCoordinates(params).then(function(response) {
                                // If the update was successful, we reload the files.
                                self.loadFiles(true);
                            },error => {
                                appCore.showError('setGps',error);
                            });
                        }
                        else if (undo.type == 'dest') {
                            let lonlat = ol.proj.transform(diFilesLMF.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
                            let f = getPrimaryFeature(diFilesLMF);
                            
                            undo.id = f.get('file').id;

                            IMatch.setGPSCoordinates({
                                target: 'dest',
                                destlng: lonlat[0],
                                destlat: lonlat[1],
                                id: f.get('file').id
                            }).then(function(response) {
                                // If the update was successful, we reload the files.
                                self.loadFiles(true);
                            },error => {
                                appCore.showError('setGps',error);
                            });
                        }

                        appCore.setUndo(true);

                    },diFiles);

                    if (showFileRoute) {
                        self.loadFileRoute();
                    }
                    
                    appCore.checkCameraDB();

                    createFOV();
                }
            });
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function setTargetMarker(lon,lat)
    {
        if (markerTarget == null) {
            // Create the feature
            var options = {
                featureType: 'target',
                geometry: new ol.geom.Point(ol.proj.fromLonLat([lon,lat])),
            };

            markerTarget = new ol.Feature(options);        
            markerTarget.setId(0);

            markerSource.addFeature(markerTarget);

            // Setup the interaction to handle drag of the target marker.
            if (diTarget) {
                self.olMap.removeInteraction(diTarget);
            }
            diTarget = new ol.interaction.Modify({
                features: new ol.Collection([markerTarget]),
                style: null,
                pixelTolerance: 20
            });
            self.olMap.addInteraction(diTarget);

            // Add the event to the drag and drop feature
            diTarget.on('modifyend',function(e) {
            },markerTarget);

            appCore.onTargetMarkerChanged(true);
        }
        else {
            markerTarget.setGeometry(new ol.geom.Point( ol.proj.fromLonLat([lon, lat])));
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Undo the last operation
    this.onUndo = function() {
        if (undo.type == 'location') {
            IMWS.post('v1/geolocation/update',{
                guid: undo.location.guid,
                lng: undo.location.lng,
                lat: undo.location.lat,
                radius: undo.location.mapRadius
            }).then(function(response) {
                if (response.result != 'ok') {
                    loadLocations();
                }
            });
        }
        else if (undo.type == 'file') {
            IMatch.setGPSCoordinates({
                target: 'created',
                lng: undo.gps.lon,
                lat: undo.gps.lat,
                id: undo.id
            }).then(function(response) {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });
        }
        else if (undo.type == 'dest') {
            IMatch.setGPSCoordinates({
                target: 'dest',
                destlng: undo.gps.destLon,
                destlat: undo.gps.destLat,
                id: undo.id
            }).then(function(response) {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });
        }
        clearUndo();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Show all files
    this.onZoomToExtent = function() {
        if (self.olMap && vectorSource) {
            var extent = vectorSource.getExtent();
            self.olMap.getView().fit(extent, self.olMap.getSize());
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Let the user draw a selection
    this.onSelectTool = function()
    {
        closePopup();
        removeSelectTool();

        appCore.switchToolbar('select');

        if (selectTool.source == null) {
            selectTool.source = new ol.source.Vector({
                features: new ol.Collection()
            });

            selectTool.layer = new ol.layer.Vector(
            {
                source: selectTool.source,
                style: selectTool.style
            });

            self.olMap.addLayer(selectTool.layer);
        }

        // Remember
        selectTool.objectType = appCore.appConfig.selectTool;

        var centerTo = null;

        var extent = self.olMap.getView().calculateExtent(self.olMap.getSize());
        var w = (extent[2] - extent[0]);
        var center = self.olMap.getView().getCenter();

        if (appCore.appConfig.selectTool == 'circle') {
            if (selectTool.circle !== undefined) {
                // Use the same radius as before, but place to the current map center
                selectTool.object = new ol.geom.Circle(center, selectTool.circle.radius);
            }
            else {
                // A new circle at 1/10 of the visible width.
                selectTool.object = new ol.geom.Circle(center, w / 10);
            }

            centerTo = selectTool.object.getCenter();
        }
        else if (appCore.appConfig.selectTool == 'rectangle') {
            /* Currently not supported (not offered in the menu for OL Maps))
            if (selectTool.rect !== undefined) {
                // Use the same dimensions as before, but shift to the current map center
                var dy = selectTool.mapCenter[0] - center[0];
                var dx = selectTool.mapCenter[1] - center[1];
                var e = selectTool.rect.extent;
                e[0] += dx;
                e[2] += dx;
                e[1] += dy;
                e[3] += dy;
                selectTool.object = new ol.geom.Polygon.fromExtent(e);
            }
            else {
                var wh = w/10;
                var e = [center[0] - wh, center[1] - wh, center[0] + wh, center[1]+wh];
                selectTool.object = new ol.geom.Polygon.fromExtent(e);
            }
            centerTo = selectTool.object.getCenter();
            */
        }
        else if (appCore.appConfig.selectTool == 'polygon') {
            if (selectTool.poly !== undefined) {
                // Use the same coordinates as before, but shift to the new center
                var dy = selectTool.mapCenter[0] - center[0];
                var dx = selectTool.mapCenter[1] - center[1];

                var coord = selectTool.poly.coord[0];

                for (var i = 0; i < coord.length; ++i) {
                    coord[i][0] -= dy;
                    coord[i][1] -= dx;
                }

                selectTool.object = new ol.geom.Polygon([coord],'XY');
            }
            else {
                var wh = w/10;
                var coord = [
                    [center[0] - wh, center[1] + wh],
                    [center[0] + wh, center[1] + wh],
                    [center[0], center[1] - wh],
                    [center[0] - wh, center[1] + wh]
                ]
                selectTool.object = new ol.geom.Polygon([coord],'XY');
            }
            centerTo = center;
        }

        selectTool.mapCenter = self.olMap.getView().getCenter();

        if (selectTool.object) {
            selectTool.source.addFeatures([ new ol.Feature({ geometry: selectTool.object }) ]);

            var fc = new ol.Collection(selectTool.source.getFeatures());

            // Allow the user to resize the objects, add new vertices.
            selectTool.iaModify = new ol.interaction.Modify({
                features: fc,
                pixelTolerance: 20
            });
            self.olMap.addInteraction(selectTool.iaModify);

            // Allow the user to drag the objects around
            selectTool.iaTranslate = new ol.interaction.Translate({
                features: fc,
                pixelTolerance: 20
            });
            self.olMap.addInteraction(selectTool.iaTranslate);

            // Add the event to the drag and drop feature
            selectTool.object.on('change',function(e) {
                selectTool.mapCenter = self.olMap.getView().getCenter();
            });
            
            if (centerTo != null) {
                self.olMap.getView().setCenter(centerTo);
            }        
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // 
    this.onSelectToolApply = function(apply)
    {
        if (apply) {
            // Get the coordinates of the shape and create the call for IMWS

            var params = {
            }

            // If we work with the file window, set the scope to all files in the fw.
            if (appCore.appConfig.selectScope == 'scope') {
                params.idlist = IMatch.idlist.fileWindowFiles;
            }

            // the options parameter looks different for each shape type
            var options = {
                type: selectTool.objectType,
            }

            if (selectTool.objectType == 'circle') {
                var lonlat = ol.proj.transform(selectTool.object.getCenter(), 'EPSG:3857', 'EPSG:4326');
                // Needs a center coordinate and a radius
                options.center = {
                    lat: lonlat[1],
                    lng: lonlat[0],
                }
                // We need the radius in meters, which requires a bit of math
                // See: http://stackoverflow.com/questions/32202944/openlayers-3-circle-radius-in-meters for an explanation
                var center = selectTool.object.getCenter()
                var radius = selectTool.object.getRadius()
                var edgeCoordinate = [center[0] + radius, center[1]];
                // var wgs84Sphere = new ol.Sphere(6378137);
                // options.radius = wgs84Sphere.haversineDistance(
                //     ol.proj.transform(center, 'EPSG:3857', 'EPSG:4326'), 
                //     ol.proj.transform(edgeCoordinate, 'EPSG:3857', 'EPSG:4326')
                // );                
                options.radius = ol.sphere.getDistance(ol.proj.transform(center, 'EPSG:3857', 'EPSG:4326'), ol.proj.transform(edgeCoordinate, 'EPSG:3857', 'EPSG:4326'));
            }
            else if (selectTool.objectType == 'rectangle') {
                // Needs the northEast and southWest parameters.
                // Not supported yet
            }
            else if (selectTool.objectType == 'polygon') {
                // We need all points in the options.points array
                options.points = [];
                var coord = selectTool.object.getCoordinates()[0];
                for (var i = 0; i < coord.length; ++i) {
                    var lonlat = ol.proj.transform(coord[i], 'EPSG:3857', 'EPSG:4326');
                    options.points.push({
                        lat: lonlat[1], 
                        lng: lonlat[0]
                    })
                }
            }

            params.options = JSON.stringify(options);

            IMWS.get('v1/search/geoarea',params).then(function(response) {
                console.log(JSON.stringify(response,null,2));

                if (response.result == 'ok') {

                    if (response.id.length == 0) {
                        // If we did find nothing, tell the user.
                        var m = new IMatchModals;
                        m.showMessageBoxOK({
                            headline: IMatchTranslate.getStr('MSGBOX_SELECT_NO_RESULT_TITLE'),
                            content: IMatchTranslate.getStr('MSGBOX_SELECT_NO_RESULT_TEST')
                        });
                    }
                    else {
                        if (appCore.appConfig.selectScope == 'scope') {
                            // If we have searched the scope, we select all files found.
                            IMatch.selectFiles({
                                id: response.id.join(',')
                            });
                        }
                        else {
                            // Else we open the results in a result window.
                            if (response.id.length > 0) {
                                var files = response.id.join(',');
                                IMatch.openResultWindow({
                                    name: IMatchTranslate.getStr('TB_SELECT_RESULT_WINDOW'),
                                    id: files
                                }).then(function(response) {
                                    // After the result window is open, select all files.
                                    IMatch.selectFiles({
                                        id: files
                                    });
                                });
                            }
                        }
                    }
                }
            });
        }

        removeSelectTool();
    }

    /////////////////////////////////////////////////////////////////////////////
    // 
    function removeSelectTool()
    {
        if (selectTool.object != null) {

            self.olMap.removeInteraction(selectTool.iaModify);
            selectTool.iaModify = null;

            self.olMap.removeInteraction(selectTool.iaTranslate);
            selectTool.iaTranslate = null;
            

            // We remember the last coordinate set for the shape for this session.
            if (selectTool.objectType == 'circle') {
                selectTool.circle = {}
                selectTool.circle.center = selectTool.object.getCenter();
                selectTool.circle.radius = selectTool.object.getRadius();
            }
            else if (selectTool.objectType == 'rectangle') {
                selectTool.rect = {}
                selectTool.rect.extent = selectTool.object.getExtent();
            }
            else if (selectTool.objectType == 'polygon') {
                selectTool.poly = {}
                selectTool.poly.coord = selectTool.object.getCoordinates();
            }

            selectTool.source.clear(true);
            selectTool.object = null;
            selectTool.objectType = '';
        }

        appCore.switchToolbar('default');        
    }


    /////////////////////////////////////////////////////////////////////////////
    // Center to the target marker
    this.onLocateTargetMarker = function() {
        if (markerTarget) {
            self.olMap.getView().setCenter(markerTarget.getGeometry().getCoordinates());
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Goto the location with the given GUID
    this.onGotoLocation = function(guid) {
        if (locations) {
            for (var i = 0; i < locations.length; ++i) {
                if (locations[i].guid == guid) {
                    var coord = ol.proj.fromLonLat([locations[i].lng,locations[i].lat]);
                    locationLayer.setVisible(true);
                    self.olMap.getView().setCenter(coord);
                    break;
                }
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // appCore has determined that a file was changed. 
    // We check if the file is currently on the map and update it.
    this.onFileUpdate = function(msg) {

        var hasfile = false;
        for (var i = 0; i < files.length; ++i) {
            if (files[i].id == msg.d1) {
                hasfile = true;
                break;
            }
        }

        if (hasfile) {
            loadFilesDelayed(true);
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Locations in the database were changed (by this app or something else).
    this.onLocationUpdate = function(msg)
    {
        loadLocations();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Called when the state of the lock option changes
    this.onLockChanged = function()
    {
        if (diLocations) {
            self.olMap.removeInteraction(diLocations);
        }
        if (diFiles) {
            self.olMap.removeInteraction(diFiles);
        }

        if (!appCore.appConfig.lockMove) {
            if (diLocations) {
                self.olMap.addInteraction(diLocations);
            }

            if (!appCore.dbReadOnly && diFiles) {
                self.olMap.addInteraction(diFiles);
            }
        }
    }


   /////////////////////////////////////////////////////////////////////////////
   // A track file was loaded.
   this.onTrackLoaded = function(fitTrack)
   {
        clearTrackLogs();
        trackSource.clear();        

        if (!appCore.tracks().hasTracks()) return;

        var trackData = appCore.tracks().trackData();
        var features = [];
        
        for (var i = 0; i < trackData.tracks.length; ++i) {
            var path = [];

            for (var p = 0; p < trackData.tracks[i].points.length; ++p) {
                path.push(ol.proj.fromLonLat([trackData.tracks[i].points[p].lon,trackData.tracks[i].points[p].lat]));
            }

            var poly = new ol.Feature({
                geometry: new ol.geom.LineString(path,'XY')
            });     

            poly.setStyle(new ol.style.Style({
                stroke: new ol.style.Stroke({
                    width: 6,
                    color: TRACK_COLORS[i % TRACK_COLORS_COUNT]
                })
            }));

            features.push(poly);
        }

        if (features.length > 0) {
            trackSource.addFeatures(features);
            if (fitTrack) {
                self.onCenterTrack();
            }
        }
   }


   /////////////////////////////////////////////////////////////////////////////
   // Center to the bounding box of the track.
   this.onCenterTrack = function()
   {
        if (self.olMap) {
            var extent = trackSource.getExtent();
            // We scale by 10% to make some extra space between the map 
            // borders and the visible points.
            var g = ol.geom.Polygon.fromExtent(extent)
            g.scale(1.1);
            self.olMap.getView().fit(g, self.olMap.getSize());
        }
   }


   /////////////////////////////////////////////////////////////////////////////
   // Show files contained in the current track. These files have no GPS data
   // set, the GPS data only exists in the file objects maintained by the track.
   this.onShowTrackFiles = function()
   {
       clearTrackMarkers();

       var features = [];
       appCore.tracks().trackData().files.forEach(function(f) {

            if (f.mapped) {

                // Create the feature
                var options = {
                    id: f.id,
                    file: f,
                    name: f.name,
                    title: f.mappedGPS.lat.toFixed(6) + ', ' + f.mappedGPS.lng.toFixed(6) + '\n' + (f.mappedGPS.alt).toFixed(2) + 'm\n' + moment(f.mappedGPS.time).format(),
                    geometry: new ol.geom.Point(ol.proj.fromLonLat([f.mappedGPS.lng, f.mappedGPS.lat]))
                };

                var marker = new ol.Feature(options);
                features.push(marker);
            }
        }, this);

        if (features.length > 0) {
            trackMarkerSource.addFeatures(features);
        }
    }

   

   /////////////////////////////////////////////////////////////////////////////
   // Remove track markers
   this.onHideTrackFiles = function()
   {
       clearTrackMarkers();
   }
   
   /////////////////////////////////////////////////////////////////////////////
   // A track file was unloaded
   this.onTrackUnLoaded = function()
   {
       clearTrackLogs();
   }
    
   // Remove all track markers from the map
   function clearTrackMarkers()
   {
        trackMarkerSource.clear();
   }


   // Remove the track polyline and markers from the map
   function clearTrackLogs()
   {
       clearTrackMarkers();
       trackSource.clear();       
   }


   /////////////////////////////////////////////////////////////////////////////
   // Create a track from the files loaded.
   self.enableShowFileRoute = function(show)
   {  
        showFileRoute = show;
        self.loadFileRoute();
   }

   /////////////////////////////////////////////////////////////////////////////
   //
   this.loadFileRoute = function()
   {
        clearfileRoute();
        if (!showFileRoute || !files || (files.length == 0)) {
            return;
        }            

        var features = [];
        var path = [];
       
        for (let f of files) {
            if (f.gps) {      
                path.push(ol.proj.fromLonLat([f.gps.lon,f.gps.lat]));
            }
        }

        var poly = new ol.Feature({
            geometry: new ol.geom.LineString(path,'XY')
        });     

        features.push(poly);

        if (features.length > 0) {
            fileRouteSource.addFeatures(features);
            // var extent = fileRouteSource.getExtent();
            // // We scale by 10% to make some extra space between the map 
            // // borders and the visible points.
            // var g = ol.geom.Polygon.fromExtent(extent)
            // g.scale(1.1);
            // self.olMap.getView().fit(g, self.olMap.getSize());
        }
   }

   /////////////////////////////////////////////////////////////////////////////
   // 
   function clearfileRoute()
   {
       if (fileRouteSource) {
           fileRouteSource.clear();
       }
   }

   /////////////////////////////////////////////////////////////////////////////
   // Close open popups
   this.onClosePopups = function()
   {
       closePopup();
   }


   /////////////////////////////////////////////////////////////////////////////
   // This is called by the appCore when it destroys the map
   this.onDestroy = function()
   {
        saveConfig();
   }

    /////////////////////////////////////////////////////////////////////////////
    // Update user settings and save back
    function saveConfig() {
        if (self.olMap && self.olMap.getView()) {
            appCore.appConfig.zoomLevel = self.olMap.getView().getZoom();

            var lonlat = ol.proj.transform(self.olMap.getView().getCenter(), 'EPSG:3857', 'EPSG:4326');

            appCore.appConfig.mapCenter.lng = lonlat[0];
            appCore.appConfig.mapCenter.lat = lonlat[1];

            appCore.saveConfig();
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Called by appCore when the user drops IMatch files onto the map.
    this.onDrop = function(e)
    {
        var coord = self.olMap.getCoordinateFromPixel([e.offsetX,e.offsetY]);
        var lonlat = ol.proj.transform(coord, 'EPSG:3857', 'EPSG:4326');

        clearUndo();
        IMatch.setGPSCoordinates({
            target: 'dest',
            destlng: lonlat[0],
            destlat: lonlat[1],
            idlist: IMatch.idlist.fileWindowSelection
        }).then(function(response) {
            // If the update was successful, we reload the files.
            self.loadFiles(true);
        },error => {
            appCore.showError('setGps',error);
        });
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Called by appCore when the user clicks on the appy target marker command.
    this.onApplyTargetMarker = function()
    {
        if (markerTarget) {
            function apply()
            {
                clearUndo();
                var lonlat = ol.proj.transform(markerTarget.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
                IMatch.setGPSCoordinates({
                    target: 'dest',
                    destlng: lonlat[0],
                    destlat: lonlat[1],
                    idlist: IMatch.idlist.fileWindowSelection
                }).then(function(response) {
                    // If the update was successful, we reload the files.
                    self.loadFiles(true);
                },error => {
                    appCore.showError('setGps',error);
                });
            }
            apply();
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Apply the coordinates and data of a location
    this.onApplyLocation = function(guid)
    {
        // Ask the user
        var m = new IMatchModals();
        m.showMessageBoxYesNo({
            headline: IMatchTranslate.getStr('MSGBOX_APPLY_LOC_TITLE'),
            content: IMatchTranslate.getStr('MSGBOX_APPLY_LOC_TEXT'),
            okCallback: function() {
                clearUndo();
                IMWS.post('v1/geolocation/apply',{
                    guid: guid,
                    idlist: IMatch.idlist.fileWindowSelection
                }).then(function(response) {
                    self.loadFiles(true);
                });
            }
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    this.onFindNearbyLocation = function(guid) {
        if (locations) {
            for (var i = 0; i < locations.length; ++i) {
                if (locations[i].guid == guid) {
                    IMatch.findNearbyFiles({
                        lat: locations[i].lat,
                        lng: locations[i].lng,
                        alt: 0                
                    }).then(function(response) {
                        if (response.result == 'ok') {
                            appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
                        }
                    });
                    break;
                }
            }
        }
    }
    

    /////////////////////////////////////////////////////////////////////////////
    // Position the target marker on the coordinates specified in text form.
    this.positionTargetMarker = function (lng,lat)
    {
        var coord = [lng,lat].map(Number);
        self.olMap.getView().setCenter(ol.proj.fromLonLat(coord));
        setTargetMarker(coord[0],coord[1]);
    }


    /////////////////////////////////////////////////////////////////////////////
    // Center the map at the given coordinates
    this.centerMap = function(lng,lat)
    {
       self.olMap.getView().setCenter(ol.proj.fromLonLat([lng,lat].map(Number)));
    }

    /////////////////////////////////////////////////////////////////////////////
    //
    function setFocusedMarker(feature,updateFileWindow)
    {
        if (focusedMarker == feature) return;
        updateFileWindow = updateFileWindow || false;

        if (focusedMarker != null) {
            focusedMarker.set('focused',false);
            let id = focusedMarker.get('id');
            // Force a refresh of the style for this.
            delete styleCache[id];
        }

        if (feature == null) return;

        let f = getPrimaryFeature(feature);

        focusedMarker = f;
        if (focusedMarker != null) {
            focusedMarker.set('focused',true);
            let id = focusedMarker.get('id');
            delete styleCache[id];
            if (updateFileWindow) {
                window.setTimeout(function() {
                    IMatch.focusFile({
                        id: id
                    });
                },100);
            }
            createFOV();
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function rotatePolygon(coords,angle)
    {
        var origin = coords[0];

        for (let i = 1; i < coords.length-1; ++i) {
            coords[i] = MapUtility.rotatePointArray(coords[i],origin,angle);
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function createFOV()
    {
        if (fovLayer.source != null) {
            fovLayer.source.clear(true);
        }
        else {
            fovLayer.source = new ol.source.Vector({
                features: new ol.Collection()
            });

            fovLayer.layer = new ol.layer.Vector(
            {
                source: fovLayer.source,
                style: fovLayer.style
            });

            self.olMap.addLayer(fovLayer.layer);
        }

        if (appCore.appConfig.fieldOfView == 'off') return;

        let focusid = 0;
        if (focusedMarker) {
            focusid = focusedMarker.get('id');
        }
    
        vectorSource.forEachFeature(function(feature, layer) {  
            let fid = feature.get('id');

            // If we show all FOVs or this is the focused marker
            if ((appCore.appConfig.fieldOfView == 'all') || (fid == focusid)) {
                let file = feature.get('file');
                var pmarker = LatLon(file.gps.lat, file.gps.lon);

                // We can generate a FOV when we have gos.imgDirection, with a fallback to gps.track.
                let dir = file.gps.imgDirection;
                if (dir == undefined) {
                    dir = file.gps.track;
                }
                if (dir == undefined) {
                    // If we have a destination point, we can use this as a 2nd fallback
                    if (file.gps.lat && file.gps.lon && file.gps.destLat && file.gps.destLon) {
                        dir = pmarker.bearingTo(LatLon(file.gps.destLat, file.gps.destLon));
                    }
                }
                
                if (dir != undefined) {

                    let cropFactor = 1.0;

                    // We need a make/model for a proper crop factor
                    if (file.make && file.model) {
                        let camInfo = appCore.camDB.getCameraInfo(file.make,file.model);
                        if (camInfo.result == 'ok') {
                            cropFactor = camInfo.cropFactor;
                        }
                        else if (camInfo.result != 'unknown') {
                            // This model has not set been tested.
                            appCore.camDB.lookupCameraInfo(file.make,file.model).then(response => {
                                if (response.result == 'ok') {
                                    // The model has been found.
                                    createFOV();
                                }
                            });
                        }
                    }

                    // We need one of the focal length to create the FOV
                    if (file.focalLength || file.focallength35mm) {
                        let flen35 = undefined;
                        if (parseFloat(file.focallength35mm) > 0) {
                            flen35 = parseFloat(file.focallength35mm);
                        }
                        let fov = MapUtility.getFOV(cropFactor,file.focalLength,flen35);

                        var extent = self.olMap.getView().calculateExtent(self.olMap.getSize());
                        var w = (extent[2] - extent[0]);
                        let fovSize = Math.min(100000,w*2);

                        var pdest1 = pmarker.destinationPoint(fovSize, -fov/2);
                        var pdest2 = pmarker.destinationPoint(fovSize, +fov/2);

                        var coords = [];
                        coords.push(ol.proj.fromLonLat([pmarker.lon, pmarker.lat]));
                        coords.push(ol.proj.fromLonLat([pdest1.lon, pdest1.lat]));
                        coords.push(ol.proj.fromLonLat([pdest2.lon, pdest2.lat]));
                        coords.push(coords[0]);
                        
                        rotatePolygon(coords,-dir);  // Negate Angle!

                        var fovPoly = new ol.geom.Polygon([coords],'XY');
                        fovLayer.source.addFeatures([ new ol.Feature({ geometry: fovPoly }) ]);
                    }
                }
            }

            // Continue with next
            return false;
        });
    }

    /////////////////////////////////////////////////////////////////////////////
    // Creates a destination coordinate pair for the file in the given marker.
    function createDirection(feature) {
        var lonlat = ol.proj.transform(feature.getGeometry().getCoordinates(), 'EPSG:3857', 'EPSG:4326');
        var file = feature.get('file');

        // We place the marker at 1/3 of the displayed map area
        var extent = self.olMap.getView().calculateExtent(self.olMap.getSize());
        var box = ol.proj.transformExtent(extent,'EPSG:3857','EPSG:4326'); // this looked like what I expected
        let dist = Math.abs(box[3] - box[1]) * 0.33;
        
        MapUtility.createDest(lonlat[0],lonlat[1],file, dist).then(response => {
            self.loadFiles(true);
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    // Creates a destination marker for the specified parent feature
    function createDestinationMarker(parentFeature)
    {
        // Remove direction first if it exists
        let dest = parentFeature.get('dest');
        if (dest != undefined) {
            vectorSource.removeFeature(dest.marker);
            lineSource.removeFeature(dest.line);
            parentFeature.unset('dest');
        }

        let file = parentFeature.get('file');
        
        if (file.gps.lat && file.gps.lon && file.gps.destLat && file.gps.destLon) {
            
            var marker = new ol.Feature({
                id: 'dm'+file.id,
                gps: file.gps,
                featureType: 'dest',
                name: file.name, // Used by clusters
                file: file,
                // We remember our source feature
                parentFeature: parentFeature,
                geometry: new ol.geom.Point(ol.proj.fromLonLat([file.gps.destLon, file.gps.destLat])),
            });

            // Keep track of the last changed feature
            marker.on('change',function(e) {
                diFilesLMF = e.target;
            });
            
            vectorSource.addFeature(marker);

            // Add the line between the points
            var line = new ol.Feature({
                geometry: new ol.geom.LineString([ol.proj.fromLonLat([file.gps.lon, file.gps.lat]), ol.proj.fromLonLat([file.gps.destLon, file.gps.destLat])])
            });
            lineSource.addFeature(line);

            parentFeature.set('dest',{
                marker: marker, 
                line: line
            });
        }
    }
   
    
}

