'use strict';

/*
    Collection of helper functions
*/

/* global */

var MapUtility = new function() {

    var self = this;

    /**
     * Calculates the Field of View in ° for the given crop factor and focalLength
     * @param {Float} cropFactor 
     * @param {Float} focalLength 
     */
    self.getFOV = function(cropFactor,focalLength,focalLength35mm = undefined) {

        let flen = focalLength35mm;
        if (flen == undefined) {
            flen = focalLength * cropFactor;
        }
        
        // Standard full-frame sensor => this sensor
        // 35mm camera has a sensor width of 36mm.
        // See: https://shuttermuse.com/calculate-field-of-view-camera-lens/
        return (2 * Math.atan(36 / (2 * flen))) * 180.0 / Math.PI;
    }

    /**
     * @param {*} file 
     * @returns Returns the source coordinates if available, else the dest coordinates. 
     * Returns undefined if the file has no suitable GPS coordinates.
     */
    self.getSafeCoordinates = function(file)
    {
        if (file.gps == undefined) {
            return undefined;
        }

        if ((file.gps.lon != undefined) && (file.gps.lat != undefined) && ((file.gps.lon != 0) || (file.gps.lat != 0))) {
            return {
                type: 'created',
                lon: file.gps.lon,
                lat: file.gps.lat,
            }
        }
        else if ((file.gps.destLon != undefined) && (file.gps.destLat != undefined)) {
            return {
                type: 'dest',
                lon: file.gps.destLon,
                lat: file.gps.destLat
            }
        }
        else {
            return undefined;
        }
    }

    /**
     * Updates the ImgDirecton and ImgDirectonRef for the specified file.
     * @param {*} id File id
     * @param {*} direction Angle between 0 and 359.999999°
     * @param {*} ref 'T' for true north or 'M' for magnetic north
     * @returns promise
     */
    self.setImgDirection = function(id, direction, ref)
    {
        let tasks = [
            {
                id: id,
                tag: 'gpsimgdirection',
                op: 'set',
                value: direction
            },
            {
                id: id,
                tag: 'gpsimgdirectionref',
                op: 'set',
                value: ref,
                rawvalue : true
            }
        ]

        return IMWS.post('v1/metadata',{
            tasks: JSON.stringify(tasks)
        });
    }

    /**
     * Creates a new destination coordinate pair for the given file.
     * @param {*} lat Camera location lat
     * @param {*} lon Camera location lon
     * @param {*} file The file
     * @param {*} dist The distance to the target marker
     * @returns promise
     */
    self.createDest = function(lng, lat, file, dist)
    {
        var dlat;
        if (dist) {
            dlat = lat + dist;  // about 50 m
        }
        else {
            dlat = lat + 0.0005;  // about 50 m
        }

        var dlng = lng;

        /*
            We set both coordinates.
            This ensures that reverse geo-coding (if enabled) is performed
            on both the created and the new dest address, basically moving
            the current "shown" address into created and the new dest address
            into "shown".
        */

        return IMatch.setGPSCoordinates({
            id: file.id,
            target: 'both',
            lat: lat,
            lng: lng,
            destlat: dlat,
            destlng: dlng
        });
    }

    /**
     * Rotates the point around origin by angle degrees
     * @param {Point} point (x,y)
     * @param {Point} origin 
     * @param {float} angle 
     */
    self.rotatePoint = function(point, origin, angle) {
        var angleRad = angle * Math.PI / 180.0;
        return {
            x: Math.cos(angleRad) * (point.x - origin.x) - Math.sin(angleRad) * (point.y - origin.y) + origin.x,
            y: Math.sin(angleRad) * (point.x - origin.x) + Math.cos(angleRad) * (point.y - origin.y) + origin.y
        };
    }
    
    /**
     * Rotates the point around origin by angle degrees
     * @param {Array} point [x,y]
     * @param {Array} origin [x,y]
     * @param {float} angle 
     * @returns {Array} point [x,y]
     */
    self.rotatePointArray = function(point, origin, angle) {
        var angleRad = angle * Math.PI / 180.0;
        return [
            Math.cos(angleRad) * (point[0] - origin[0]) - Math.sin(angleRad) * (point[1] - origin[1]) + origin[0],
            Math.sin(angleRad) * (point[0] - origin[0]) + Math.cos(angleRad) * (point[1] - origin[1]) + origin[1]
        ]
    }
}
