'use strict';

// This module handles the Track Log configuration

/* global appCore,moment */
function TrackLogConfig() {
    var self = this;

    self.scopeTimeline = {
        first: null,
        last: null,
    }
    self.trackTimeline = {
        first: null,
        last: null,
    }

    self.trackTimelineOriginalFirst = null;
    self.scopeTimelineOriginalFirst = null;

    // We use a local copy so we can update the track log info when the
    // user plays with the controls.
    var timeZoneShift = appCore.appConfig.trackLog.trackTimeShift;

    self.open = function()
    {
        appCore.setActiveOverlay('tracklog-config');

        $.get('track/track-config-tpl.html', function(data) {

            // Add template to DOM
            $('#app-container').append(data);

            // Translate it
            IMatchTranslate.translateElement('#tracklog-config-container');

            // Init the controls

            $('#tz-offset').val(timeZoneShift);

            $('#threshold').val(appCore.appConfig.trackLog.timeThresholdInSeconds === undefined ? 300 : appCore.appConfig.trackLog.timeThresholdInSeconds);

            $('#check-ignore-gps').prop('checked',appCore.appConfig.trackLog.ignoreFilesWithGPS);

            $('#check-remember-folder').prop('checked',appCore.appConfig.trackLog.rememberFolder);

            if (appCore.appConfig.showTrackInfoText) {
                $('#info-display').show();
            }
            else {
                $('#info-display').hide();
            }
    
            // Load the info about the current scope
            self.loadScopeInfo();

            $('#container-close').click(function(e) {
                self.close(e);
            });

            $('#btn-cancel').click(function(e) {
                self.close(e);
            });

            $('#toggle-info-display').click(function(e) {
                appCore.appConfig.showTrackInfoText = !appCore.appConfig.showTrackInfoText;
                if (appCore.appConfig.showTrackInfoText) {
                    $('#info-display').show();
                }
                else {
                    $('#info-display').hide();
                }
                appCore.saveConfig();
            });


            $(document).keydown(function(e) {
                if (e.which === 27) { // ESC
                    self.close(e);
                }
            });

            // Save settings and close
            $('#btn-ok').click(function(e) {
                appCore.appConfig.trackLog.trackTimeShift = $('#tz-offset').val();
                appCore.appConfig.trackLog.timeThresholdInSeconds = $('#threshold').val();
                appCore.appConfig.trackLog.ignoreFilesWithGPS = $('#check-ignore-gps').prop('checked');
                appCore.appConfig.trackLog.rememberFolder = $('#check-remember-folder').prop('checked');
                appCore.saveConfig();
                appCore.showTrackFiles();
                self.close(e);
            });

            // On click, when a track is loaded, adjust the tz-offset so the track
            // matches the earliest file.
            $('#btn-track-files-auto').click(() => {
                if (self.scopeTimelineOriginalFirst && self.trackTimelineOriginalFirst) {
                    if (timeZoneShift !== 0) {
                        timeZoneShift = 0;
                        loadTrackInfoDB();
                    }
                    // Diff in minutes
                    const d = self.scopeTimelineOriginalFirst.diff(self.trackTimelineOriginalFirst,'minutes') / 60;
                    // Round to 1/2 hours
                    timeZoneShift = Math.floor((d * 2.0).toFixed() / 2.0);
                    $('#tz-offset').val(timeZoneShift);
                    loadTrackInfoDB();
                }
            });

            // Delay function for updating the track info
            var loadTrackInfoDB = IMatch.debounce(function() {
                self.loadTrackInfo();  
            },1000);

            // Handle changes in the time-zone input field
            $('#tz-offset').on('input',function(e) {
                timeZoneShift = $('#tz-offset').val();
                loadTrackInfoDB(); 
            });

            // Handle changes in the threshold input field
            $('#threshold').on('input',function(e) {
                loadTrackInfoDB(); 
            });

            $('#check-ignore-gps').on('input',function(e) {
                loadTrackInfoDB(); 
            });
           

            // Roll in the container
            $('#tracklog-config-container').css('width','100%');
        });
    }

    // Close the overlay
    self.close = function(e)
    {
        $('#tracklog-config-container').css('width','0%');
        $('#tracklog-config-container').remove();
        
        appCore.setActiveOverlay('');
       
        if (e !== undefined) {
            e.preventDefault();
        }
        return false;
    }

    // Retrieves info about the files in the active file window and calculates
    // the earliest and latest timestamps.
    self.loadScopeInfo = function()
    {
        let self = this;

        IMWS.get('v1/files',{
            idlist: IMatch.idlist.fileWindowSelection,
            fields: 'id',
            tagdateTime: 'createdate',
            tagcreateDate: 'datecreated'
        }).then(function(response) {

            self.scopeTimeline.first = null;
            self.scopeTimeline.last = null;

            response.files.forEach(function(f) {

                var d = f.dateTime;
                if (d == '') d = f.createDate;
                if (d !== '') {
                    var m = moment(d,['YYYY:MM:DD HH:mm:ssZZ','YYYY-MM-DD HH:mm:ssZZ','YYYY:MM:DDTHH:mm:ssZZ','YYYY-MM-DDTHH:mm:ssZZ']);
                    if (m.isValid()) {
                        if (self.scopeTimeline.first == null) {
                            self.scopeTimeline.first = m;
                        }
                        else if (m < self.scopeTimeline.first) {
                            self.scopeTimeline.first = m;
                        }
                        if (self.scopeTimeline.last == null) {
                            self.scopeTimeline.last = m;
                        }
                        else {
                            if (m > self.scopeTimeline.last) {
                                self.scopeTimeline.last = m;
                            }
                        }
                    }
                }
            }, this);
            
            self.scopeTimelineOriginalFirst = new moment(self.scopeTimeline.first);

            if (self.scopeTimeline.first && self.scopeTimeline.last) {
                $('#scope-container').fadeIn();
                var msg = IMatchTranslate.getStr('TRACK_CONFIG_TZO_SCOPE');
                msg = msg.replace('{0}','<strong>' + self.scopeTimeline.first.format('L LTS') + '</strong>');
                msg = msg.replace('{1}','<strong>' + self.scopeTimeline.last.format('L LTS') + '</strong>');
                msg = msg.replace('{2}','<strong>' + self.scopeTimeline.first.utc().format('L LTS') + '</strong>');
                msg = msg.replace('{3}','<strong>' + self.scopeTimeline.last.utc().format('L LTS') + '</strong>');
                msg = msg.replace('{4}',moment.duration(self.scopeTimeline.last.diff(self.scopeTimeline.first)).asHours().toFixed(2));
                $('#scope').html(msg);
            }

            // If we have a current track, load the info as well.
            self.loadTrackInfo();
        });
    }

    // If the user has loaded a track log, we display the first and last timestamp.
    self.loadTrackInfo = function()
    {
        let self = this;

        if (appCore.tracks().hasTracks()) {

            var td = appCore.tracks().trackData();

            self.trackTimeline.first = null;
            self.trackTimeline.last = null;

            // Find the first last timestamp. These must not be necessarily in order,
            // because we also support waypoints and routes!

            for (var i = 0; i < td.tracks.length; ++i) {
                for (var p = 0; p < td.tracks[i].points.length; ++p) {
                    var m = moment.utc(td.tracks[i].points[p].time,'YYYY-MM-DD HH:mm:ss');
                    if (m.isValid()) {
                        // We remember the first unmodified timestamp
                        if (self.trackTimelineOriginalFirst == null) {
                            // DEEP copy!
                            self.trackTimelineOriginalFirst = new moment(m);
                        }
                        else if (m < self.trackTimelineOriginalFirst) {
                            self.trackTimelineOriginalFirst = new moment(m);
                        }
                       
                        m.add(timeZoneShift*3600,'s');
                        if (self.trackTimeline.first == null) {
                            self.trackTimeline.first = m;
                        }
                        else if (m < self.trackTimeline.first) {
                            self.trackTimeline.first = m;
                        }
                        if (self.trackTimeline.last == null) {
                            self.trackTimeline.last = m;
                        }
                        else {
                            if (m > self.trackTimeline.last) {
                                self.trackTimeline.last = m;
                            }
                        }
                    }
                }
            }
            
            if (self.trackTimeline.first && self.trackTimeline.last) {
                $('#no-track-container').hide();
                $('#track-container').fadeIn();

                var msg = IMatchTranslate.getStr('TRACK_CONFIG_TZO_TRACK');
                msg = msg.replace('{0}','<strong>' + self.trackTimeline.first.utc().format('L LTS') + '</strong>');
                msg = msg.replace('{1}','<strong>' + self.trackTimeline.last.utc().format('L LTS') + '</strong>');
                msg = msg.replace('{4}',moment.duration(self.trackTimeline.last.diff(self.trackTimeline.first)).asHours().toFixed(2));
                if (self.scopeTimeline.first) {
                    msg = msg.replace('{5}',moment.duration(self.trackTimeline.first.diff(self.scopeTimeline.first)).asHours().toFixed(2));
                }
                else {
                    // No files are in the scope or none has GPS data.
                    msg = msg.replace('{5}','0');
                }
                
                $('#track').html(msg);

                if (self.scopeTimeline.first) {
                    // The time axis, showing the times from the images and the times from the track as two bars.
                    // The image timeline is always 100%.
                    // The track timeline sits below, visualizing the offset and length.
                    var tl = $('#tl-track');
                    // ps = 100%
                    let ps = self.scopeTimeline.last.diff(self.scopeTimeline.first);
                    let pt = self.trackTimeline.last.diff(self.trackTimeline.first);
                    // Length of the track timeline (may be > 100%)
                    let d = Math.min(100,(pt * 100) / ps);
                    // Offset of the track timeline (may be < 0)
                    let dt = self.trackTimeline.first.diff(self.scopeTimeline.first);
                    let poffset = Math.max(0,Math.min(100,(dt*100)/ps));
                    //console.log('poffset: ' + poffset);
                    tl.css('width',Math.max(1,Math.min(100,d-poffset))+'%');
                    tl.css('margin-left',poffset+'%');

                    $('#tl-img').show();
                    $('#tl-track').show();
                }
                else {
                    $('#tl-img').hide();
                    $('#tl-track').hide();
                }
            }

            self.mapFiles();
        }
    }

    // Map the files after the time offset has changed to see how many we match
    self.mapFiles = function()
    {
        if (appCore.tracks().hasTracks()) {
            appCore.tracks().mapFiles(timeZoneShift, $('#threshold').val(), $('#check-ignore-gps').prop('checked')).then(function() {
                let s = IMatchTranslate.getStr('TRACK_CONFIG_FILES_MAPPED')
                    .replace('{0}',appCore.tracks().trackData().filesMapped)
                    .replace('{1}',Math.max(appCore.tracks().trackData().filesToMap - appCore.tracks().trackData().filesIgnored))
                    .replace('{2}',appCore.tracks().trackData().filesIgnored)
                $('#track-files-mapped').text(s);
            },error => {
                $('#track-files-mapped').text(``);
            });
        }
    }

}

