/**
 * This file contains helper functions for mapping tracks.
 * It relies on functions from the geodesy/*.js libraries.
 */
'use strict';

/* global moment,LatLon,appCore,IMatchTrackLog */

function IMatchTracks() {
    
    var self = this;

    var trackData = {
        // The currently loaded tracks
        tracks: [],
        
        // If files have been mapped, the mapped files.
        files: [],

        // The files in th elast mapFiles() call which were valid (no GPS data, valid timestamp)
        filesToMap : 0,

        // The number of files which were actually mapped. 
        // If this is different from filesToMap, some files were not mapped
        filesMapped : 0,

        // Files ae ignored when they have no valid date or GPS data
        filesIgnored : 0
    }

    // This is what we store in the file.errror
    this.ERRORS = {
        ERR_INVALID_DATE : 1
    }

    /////////////////////////////////////////////////////////////////////////////
    // Returns the current track data
    this.trackData = function()
    {
        return trackData;
    }

    /////////////////////////////////////////////////////////////////////////////
    // Returns true if tracks are loaded
    this.hasTracks = function()
    {
        return trackData.tracks.length > 0;
    }

    /////////////////////////////////////////////////////////////////////////////
    // Unload the tracks
    this.clear = function()
    {
        trackData.tracks = [];
        trackData.error = '';
    }


    /////////////////////////////////////////////////////////////////////////////
    // Loads the given GPX track log into trackData.
    // The function returns a promise that is resolved once the work is done.
    // If there is a problem with loading the file, the promise is rejected.
    this.loadTrackLog = function(fileName)
    {
        self.clear();

        var dfd = jQuery.Deferred();

        IMatch.readTextFile({
            filename: fileName
        }).then(function(response) {
            // Parse the GPX data
            var tl = new IMatchTrackLog();
            if (tl.load(response.data)) {
                // And save it trackData
                trackData.tracks = tl.tracks();
                dfd.resolve();
            }
            else {
                dfd.reject(tl.error());
            }
        },function(error) {
            // Some kind of server error
            dfd.reject(error.status);
        });

        return dfd.promise();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Loads all GPX track logs found in the given folder into tracks.
    // The function returns a promise that is resolved once the work is done.
    // If there is a problem with loading the file, the promise is rejected.
    this.loadTrackLogsFromFolder = async function(path)
    {
        self.clear();

        var dfd = jQuery.Deferred();

        try {
            trackData.tracks = [];
            const folderContents = await IMatch.folderScan({
                path: path,
                filemask: '*.gpx',
                recursive: false
            });

            if (folderContents.folders.length == 0) {
                // Nothing found.
                dfd.reject();
                return;
            }

            for (let i = 0; i < folderContents.folders[0].files.length; i++) {
                var fileData = await IMatch.readTextFile({
                     filename: folderContents.folders[0].files[i].fullName
                });
                var tl = new IMatchTrackLog();
                if (tl.load(fileData.data)) {
                    trackData.tracks = trackData.tracks.concat(tl.tracks());
                }
                else {
                    dfd.reject(tl.error());
                    return;
                }
            };

            dfd.resolve();
            return dfd.promise();
        }
        catch (error)
        {
            dfd.reject(error.status);
            return dfd.promise();
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Returns the difference in seconds between timeA and timeB.
    this.calculateDifferenceInSeconds = function(timeA, timeB)
    {
        const a = moment(timeA);
        const b = moment(timeB);
        return Math.abs(a.diff(b,'seconds'));
    }


    /////////////////////////////////////////////////////////////////////////////
    // This function uses the currently loaded track(s) to map all files in 
    // the current scope. If the function succeeds, it sets the files memmber of trackdata.
    // 
    // The function is asnychronous and returns a Promise!
    //
    // For each file, some core data and the information about the mapped coordinates is filled.
    // The 'mapped' element indicates whether or not a file was mapped successfully.
    // The mappedGPS contains the mapped GPS data and time stamp.
    this.mapFiles = function(timeOffset, maxTimeDifferenceInSeconds, ignoreFilesWithCoordinates)
    {
        var dfd = jQuery.Deferred();

        const timeShift = timeOffset === undefined ? appCore.appConfig.trackLog.trackTimeShift : timeOffset;
        const threshboldInSeconds = maxTimeDifferenceInSeconds === undefined ? appCore.appConfig.trackLog.timeThresholdInSeconds : maxTimeDifferenceInSeconds;
        const ignoreFilesWithGPS = ignoreFilesWithCoordinates === undefined ? appCore.appConfig.trackLog.ignoreFilesWithGPS : ignoreFilesWithCoordinates;

        let self = this;

        trackData.files = [];
        trackData.filesToMap = 0;
        trackData.filesMapped = 0;
        trackData.filesIgnored = 0;

        IMWS.get('v1/files',{
            idlist: IMatch.idlist.fileWindowSelection, // All selected files in the current scoope
            fields: 'id,gps,name,datetimeutc'
        }).then(function(response) {
            trackData.files = response.files;

            trackData.files.forEach(function(f) {
                f.mapped = false;
                f.valid = true;

                // May not be an image or may have no timestamp!
                // If some of the files in the scope are no images, for example.
                if (f.dateTimeUTC != '') {
                    // Timestamps in metadata may come in different formats, depending on the camera / software used.
                    // We try to parse various formats below. Typical metadata mess :-/
                    // Still, this may fail in some cases. We then mark the file as invalid and ignore it.

                    // Timestamps without a time zone are considered to be in local time!

                    var m = moment.utc(f.dateTimeUTC,['YYYY:MM:DD HH:mm:ssZZ','YYYY-MM-DD HH:mm:ssZZ','YYYY:MM:DDTHH:mm:ssZZ','YYYY-MM-DDTHH:mm:ssZZ']);
                    if (m.isValid()) {
                        f.ts = m.utc().valueOf();
                    }
                    else {
                        f.valid = false;
                        f.error = self.ERR_INVALID_DATE;
                    }
                }
                else {
                    f.valid = false;
                    f.error = self.ERR_INVALID_DATE;
                }
            }, this);

            // Prepare the GPX track points.

            for (var t = 0; t < trackData.tracks.length; ++t) {

                // Calculate the UNIX timestamp for each point from the time recorded in the point.

                var track = trackData.tracks[t];
                for (var p = 0; p < track.points.length; ++p) {
                    // The points are always in UTC.
                    // Add the time shift value for the track configured by the user.
                    track.points[p].ts = (timeShift*3600*1000) + moment.utc(track.points[p].time,'YYYY-MM-DD HH:mm:ssZZ').valueOf();
                }

                // Sort the track points by time stamp ascending

                track.points.sort(function(a,b) {
                    const r = a.ts - b.ts;
                    if (r !== 0) {
                        return r;
                    } else {
                        return a.lat < b.lat;
                    }
                });

// console.log(track.points);

                // Remember the earliest and latest ts in this track.
                // We use that to quickly weed files not on the track.
                // Extend by the threshold to catch the first and last point correctly.

                if (track.points.length > 0) {
                    track.minTs = track.maxTs = track.points[0].ts;
                    track.minTs -= (threshboldInSeconds * 1000);
                    if (track.points.length > 1) {
                        track.maxTs = track.points[track.points.length-1].ts + (threshboldInSeconds * 1000);
                    }
                }
            }

// console.log(`Track: ${track.points.length} points, min time: ${moment(track.minTs).format()}, max time: ${moment(track.maxTs).format()}`);

            // Match all files in the current scope which don't already have GPS data.
            trackData.files.forEach(function(f) {
                
                // Not mapped yet.
                f.mapped = false;
                trackData.filesToMap++;
                
                // We ignore files which are not valid and which already have GPS data (except enabled by the user).
                if (f.valid && (!ignoreFilesWithGPS || (f.gps === undefined))) {

                    const fileMoment = moment(f.ts);

                    console.log(`File [${f.name}]`);

                    // For each track
                    for (var t = 0; t < trackData.tracks.length; ++t) {
                        const track = trackData.tracks[t];

                        // Do a quick check to see if the file falls within the track window
                        if ((f.ts >= track.minTs) && (f.ts <= track.maxTs)) {

                            // Find all track points with ts within threshold of the file ts.

                            var mindiff = 999999999;
                            let selectedTrackPoint = null;
                            let selectedTrackPointIndex = 0;
                            for (var p = 0; p < track.points.length; ++p) {
                                const diff = Math.abs(fileMoment.diff(moment.utc(track.points[p].ts),'seconds'));
                                if (diff <= threshboldInSeconds) {
                                    mindiff = diff;
                                    selectedTrackPoint = track.points[p];
                                    selectedTrackPointIndex = p;

                                    // Continue with the next point until we find a point outside the threshold.
                                    for (var i = p + 1; i < track.points.length; ++i) {
                                        const diff = Math.abs(fileMoment.diff(moment.utc(track.points[i].ts),'seconds'));
                                        if (diff < mindiff) {
                                            selectedTrackPoint = track.points[i];
                                            selectedTrackPointIndex = i;
                                            mindiff = diff;
                                        }
                                        else {
                                            break;
                                        }
                                    }
                                    break;
                                }
                            }

                            if (selectedTrackPoint !== null) {
                                f.mapped = true;
                                trackData.filesMapped++;

// console.log(`Mapped file [${f.id}] '${f.name}' with time stamp ${f.dateTimeUTC} to point ${selectedTrackPointIndex} with time stamp ${selectedTrackPoint.time}, shifted to ${moment.utc(selectedTrackPoint.ts).format()}, lat: ${selectedTrackPoint.lat}, lon: ${selectedTrackPoint.lon}.`);

                                /*  
                                Found it. We can now either directly assign the coordinates 
                                of this point, or we look at the next point and interpolate
                                to get a better fix.

                                Assuming 
                                    A : The point we just found (file has been taken at or after this point)
                                    B : The point following A (at p + 1)

                                Theory
                                    f.ts is > A.ts and <= B.ts.
                                    This means that the image has been taken somewhere in the middle between A and B.
                                    Assigning f to A.coordinates is thus not truly correct.
                                    f should get the coordinates of the center point between A and B, or,  better, 
                                    a point along a virtual line between A and B that matches the time offset
                                    from f.ts to A.ts. 

                                    A       file               B
                                    |        |                 |

                                Example
                                    A.ts : 12:05:00
                                    F.ts : 12:20:00  (15 seconds after A)
                                    B.ts : 12:55:00  (50 seconds after A)

                                    Delta t B->A: 50s
                                    Delta t F->: 15s
                                    50 / 15: 0.33

                                    The coordinates for f should thus be taken from the point 15 seconds 'after' A.ts.

                                    1. We construct a line between A and B (say: 1000 meters)
                                    2. We calculate the time difference between B and A (50 seconds).
                                    3. f thus gets the coordinates of the point at at A + 1000/3m => A + 333.33m.
                                */          

                                // Only if there is a 'next' point. If p = 0 we use the point as is.
                                if ((selectedTrackPointIndex + 1) < track.points.length) {
                                    const nextTrackPoint = track.points[selectedTrackPointIndex + 1];

                                    // Time difference between A and B in ms.
                                    const dtTrackPoints = nextTrackPoint.ts - selectedTrackPoint.ts;

                                    // Time difference between file and A
                                    const dtFile = Math.min(f.ts - selectedTrackPoint.ts);

                                    // Note: This may become negative if the file ts is before the track ts!
                                    const delta = Math.max(0,Math.min(1,dtFile / Math.max(1,dtTrackPoints)));

                                    // Calculate the distance between the points in meters (using geodesy/LatNon)
                                    const pa = LatLon(selectedTrackPoint.lat,selectedTrackPoint.lon);
                                    const pb = LatLon(nextTrackPoint.lat,nextTrackPoint.lon);

                                    // Calculate the point on the line between A and B (0 - 100% = 0 - 1.0) that best matches the time of the file.
                                    var pfile;

                                    if ((pa.lat == pb.lat) && (pa.lon == pb.lon)) {
                                        // intermediatePointTo returns NaN in that case!
                                        // We use pa.
                                        pfile = pa;
                                    }
                                    else {
                                        pfile = pa.intermediatePointTo(pb, delta);
                                        // Just in case...
                                        if (isNaN(pfile.lat) || isNaN(pfile.lon)) {
                                            pfile = pa;
                                        }
                                    }

                                    // We apply the same logig to the alt component.
                                    var alt = selectedTrackPoint.alt + (nextTrackPoint.alt - selectedTrackPoint.alt) * delta;

                                    // The interpolated GPS coordinates for this file
                                    f.mappedGPS = {
                                        lat: pfile.lat,
                                        lng: pfile.lon,
                                        alt: alt,
                                        time: f.ts,
                                    }
                                }
                                else {
                                    // Just map to the first point in the track log.
                                    f.mappedGPS = {
                                        lat: selectedTrackPoint.lat,
                                        lng: selectedTrackPoint.lon,
                                        alt: selectedTrackPoint.alt,
                                        time: f.ts,
                                    }
                                }
                                // Done with this file.
                                break;
                            }
                            if (f.mapped) break;
                        }  // In track window
                    }
                }
                else {
                    trackData.filesIgnored++;
                }
            }); // trackData.files.forEach(function(f) {

            //console.log(JSON.stringify(trackData.files,null,2));
            //console.log(JSON.stringify(trackData.tracks,null,2));

            // console.log(trackData.filesMapped + ' of ' + trackData.filesToMap + ' matched.');

            // if (trackData.filesMapped < trackData.filesToMap) {
            //     console.log('Not matched:');

            //     trackData.files.forEach(function(f) {
            //         if (!f.matched) {
            //             console.log(JSON.stringify(f,null,2));
            //         }
            //     });
            // }

            dfd.resolve();

       },function(error) {
           dfd.reject(error);
       });  // files

       return dfd.promise();
    }
}